@@ -1,5 +1,5 @@
 /* events.c -- Event handling
-   $Id: events.c,v 1.111 2000/06/07 09:35:24 john Exp $
+   $Id: events.c,v 1.129 2000/11/13 23:57:43 jsh Exp $
 
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
@@ -41,6 +41,7 @@ static int current_mouse_x, current_mouse_y;
 /* ..and the position (and window) at the last button-press */
 static int button_press_mouse_x = -1, button_press_mouse_y = -1;
 static Window button_press_window;
+static bool pointer_in_motion;
 
 /* Most recently seen server timestamp. */
 Time last_event_time;
@@ -53,8 +54,6 @@ static repv current_event_window;
 /* We need a ButtonRelease on this fp. */
 struct frame_part *clicked_frame_part;
 
-static repv current_context_map;
-
 static XID event_handler_context;
 
 static Atom xa_sawmill_timestamp;
@@ -78,6 +77,9 @@ DEFSYM(shape_notify_hook, "shape-notify-hook");
 DEFSYM(enter_frame_part_hook, "enter-frame-part-hook");
 DEFSYM(leave_frame_part_hook, "leave-frame-part-hook");
 DEFSYM(configure_request_hook, "configure-request-hook");
+DEFSYM(window_state_change_hook, "window-state-change-hook");
+
+DEFSYM(pointer_motion_threshold, "pointer-motion-threshold");
 
 /* for enter/leave-notify-hook */
 DEFSYM(root, "root");
@@ -86,11 +88,25 @@ DEFSYM(root, "root");
 DEFSYM(new_value, "new-value");
 DEFSYM(deleted, "deleted");
 
+DEFSYM(urgency, "urgency");
+
 DEFSYM(stack, "stack");
 DEFSYM(above, "above");
 DEFSYM(below, "below");
+DEFSYM(top_if, "top-if");
+DEFSYM(bottom_if, "bottom-if");
+DEFSYM(opposite, "opposite");
 DEFSYM(dimensions, "dimensions");
 
+/* for enter/leave-notify-hook */
+DEFSYM(normal, "normal");
+DEFSYM(grab, "grab");
+DEFSYM(ungrab, "ungrab");
+
+DEFSYM(ignore_fp_keymap, "ignore-fp-keymap");
+
+repv Fsynthetic_configure_mutex (repv);
+
 /* `Time' will always be 32-bits, due to underlying wire protocol (?) */
 #define TIME_MAX 4294967295UL
 
@@ -155,35 +171,47 @@ record_event_time (XEvent *ev)
 static void
 record_mouse_position (int x, int y, int event_type, Window w)
 {
-    current_mouse_x = x;
-    current_mouse_y = y;
+    bool update_press = FALSE;
+
     switch (event_type)
     {
     case ButtonPress:
-	button_press_mouse_x = x;
-	button_press_mouse_y = y;
-	button_press_window = w;
+	update_press = TRUE;
 	break;
 
     case MotionNotify:
 	if (button_press_mouse_x == -1)
-	{
-	    button_press_mouse_x = x;
-	    button_press_mouse_y = y;
-	    button_press_window = w;
-	}
+	    update_press = TRUE;
 	break;
 
     case ButtonRelease:
 	if (button_press_mouse_x == -1)
-	{
-	    button_press_mouse_x = x;
-	    button_press_mouse_y = y;
-	    button_press_window = w;
-	}
+	    update_press = TRUE;
 	break;
     }
+
+    if (update_press)
+    {
+	button_press_mouse_x = x;
+	button_press_mouse_y = y;
+	button_press_window = w;
+	pointer_in_motion = FALSE;
+    }
+
+    current_mouse_x = x;
+    current_mouse_y = y;
     current_event_updated_mouse = TRUE;
+
+    if (event_type == MotionNotify && !pointer_in_motion)
+    {
+	int delta_x = x - button_press_mouse_x;
+	int delta_y = y - button_press_mouse_y;
+	repv tem = Fsymbol_value (Qpointer_motion_threshold, Qt);
+	int threshold = rep_INTP (tem) ? rep_INT (tem) : 0;
+	
+	if (ABS (delta_x) > threshold || ABS (delta_y) > threshold)
+	    pointer_in_motion = TRUE;
+    }
 }
 
 static void
@@ -295,6 +323,27 @@ synthesize_button_release (void)
     button_press_window = 0;
 }
 
+static repv
+current_context_map (void)
+{
+    repv map = Qnil;
+
+    /* Only use the context map if the frame part is currently clicked,
+       and it's window is visible (i.e. not iconified) */
+    if (clicked_frame_part
+	&& clicked_frame_part->clicked
+	&& clicked_frame_part->win != 0
+	&& clicked_frame_part->win->visible)
+    {
+	repv tem = Fwindow_get (rep_VAL (clicked_frame_part->win),
+				Qignore_fp_keymap);
+	if (tem == Qnil)
+	    map = get_keymap_for_frame_part (clicked_frame_part);
+    }
+
+    return map;
+}
+
 static void
 button_press (XEvent *ev)
 {
@@ -310,19 +359,10 @@ button_press (XEvent *ev)
 	w = fp->win;
 
 	if (ev->type == ButtonPress)
-	{
 	    handle_fp_click (fp, ev);
-	    current_context_map = get_keymap_for_frame_part (fp);
-	}
     }
 
-    /* Only use the context map if the frame part is currently clicked,
-       and it's window is visible (i.e. not iconified) */
-    eval_input_event ((clicked_frame_part
-		       && clicked_frame_part->clicked
-		       && clicked_frame_part->win != 0
-		       && clicked_frame_part->win->visible)
-		      ? current_context_map : Qnil);
+    eval_input_event (current_context_map ());
 
     if (fp != 0 && w->id != 0 && ev->type == ButtonRelease)
     {
@@ -341,7 +381,6 @@ button_press (XEvent *ev)
     {
 	button_press_mouse_x = button_press_mouse_y = -1;
 	button_press_window = 0;
-	current_context_map = Qnil;
 	/* The pointer is _always_ ungrabbed after a button-release */
 	XUngrabPointer (dpy, last_event_time);
     }
@@ -369,7 +408,8 @@ motion_notify (XEvent *ev)
 	record_mouse_position (x, y, ev->type, ev->xmotion.window);
     }
 
-    eval_input_event (current_context_map);
+    if (pointer_in_motion)
+	eval_input_event (current_context_map ());
 
     XAllowEvents (dpy, SyncPointer, last_event_time);
 }
@@ -378,9 +418,13 @@ static void
 property_notify (XEvent *ev)
 {
     Lisp_Window *w = find_window_by_id (ev->xproperty.window);
+    repv w_ = rep_VAL (w);		/* type alias for gc-pro'ing */
     if (w != 0 && ev->xproperty.window == w->id)
     {
 	bool need_refresh = FALSE, changed = TRUE;
+	repv changed_states = Qnil;
+	rep_GC_root gc_w, gc_changed_states;
+
 	switch (ev->xproperty.atom)
 	{
 	    u_char *prop;
@@ -388,6 +432,7 @@ property_notify (XEvent *ev)
 	    int format;
 	    long nitems, bytes_after;
 	    long supplied;
+	    bool old_urgency, new_urgency;
 
 	case XA_WM_NAME:
 	case XA_WM_ICON_NAME:
@@ -453,9 +498,15 @@ property_notify (XEvent *ev)
 	    break;
 
 	case XA_WM_HINTS:
+	    old_urgency = w->wmhints != 0 && w->wmhints->flags & XUrgencyHint;
 	    if (w->wmhints != 0)
 		XFree (w->wmhints);
 	    w->wmhints = XGetWMHints (dpy, w->id);
+	    new_urgency = w->wmhints != 0 && w->wmhints->flags & XUrgencyHint;
+	    w->icon_image = rep_NULL;
+	    need_refresh = TRUE;
+	    if (old_urgency != new_urgency)
+		changed_states = Fcons (Qurgency, changed_states);
 	    break;
 
 	case XA_WM_NORMAL_HINTS:
@@ -482,6 +533,9 @@ property_notify (XEvent *ev)
 	    }
 	}
 
+	rep_PUSHGC (gc_w, w_);
+	rep_PUSHGC (gc_changed_states, changed_states);
+
 	if (need_refresh && w->reparented
 	    && w->property_change != 0 && w->id != 0)
 	{
@@ -496,6 +550,13 @@ property_notify (XEvent *ev)
 					   == PropertyNewValue
 					   ? Qnew_value : Qdeleted), Qnil);
 	}
+	if (changed_states != Qnil)
+	{
+	    Fcall_window_hook (Qwindow_state_change_hook, rep_VAL (w),
+			       rep_LIST_1 (changed_states), Qnil);
+	}
+
+	rep_POPGC; rep_POPGC;
     }
 }
 
@@ -562,6 +623,7 @@ destroy_notify (XEvent *ev)
 void
 map_request (XEvent *ev)
 {
+    DEFSTRING (iconify_mod, "sawfish.wm.state.iconify");
     Window id = ev->xmaprequest.window;
     Lisp_Window *w = find_window_by_id (id);
     if (w == 0)
@@ -569,12 +631,12 @@ map_request (XEvent *ev)
 	w = add_window (id);
 	if (w == 0)
 	    return;
-	w->mapped = TRUE;
 
 	if (w->wmhints && w->wmhints->flags & StateHint
 	    && w->wmhints->initial_state == IconicState)
 	{
-	    rep_call_lisp1 (Fsymbol_value (Qiconify_window, Qt), rep_VAL(w));
+	    rep_call_lisp1 (module_symbol_value (rep_VAL (&iconify_mod),
+						 Qiconify_window), rep_VAL(w));
 	}
     }
     else
@@ -587,7 +649,8 @@ map_request (XEvent *ev)
 	    Fungrab_server ();
 	}
 	w->mapped = TRUE;
-	rep_call_lisp1 (Fsymbol_value (Quniconify_window, Qt), rep_VAL(w));
+	rep_call_lisp1 (module_symbol_value (rep_VAL (&iconify_mod),
+					     Quniconify_window), rep_VAL(w));
     }
 
     if (!w->client_unmapped)
@@ -679,12 +742,21 @@ unmap_notify (XEvent *ev)
     }
 }
 
+static inline repv
+mode_to_sym (int mode)
+{
+    return (mode == NotifyNormal ? Qnormal
+	    : mode == NotifyGrab ? Qgrab
+	    : Qungrab);
+}
+
 static void
 enter_notify (XEvent *ev)
 {
     struct frame_part *fp;
+    repv mode = mode_to_sym (ev->xcrossing.mode);
     if (ev->xcrossing.window == root_window)
-	Fcall_hook (Qenter_notify_hook, Fcons (Qroot, Qnil), Qnil);
+	Fcall_hook (Qenter_notify_hook, rep_list_2 (Qroot, mode), Qnil);
     else if ((fp = find_frame_part_by_window (ev->xcrossing.window)) != 0)
     {
 	repv tem;
@@ -707,7 +779,7 @@ enter_notify (XEvent *ev)
 	if (tem && rep_CONSP(tem) && w->id != 0)
 	{
 	    Fcall_window_hook (Qenter_frame_part_hook, rep_VAL(w),
-			       Fcons (rep_VAL(fp), Qnil), Qnil);
+			       rep_list_2 (rep_VAL(fp), mode), Qnil);
 	}
     }
     else
@@ -716,7 +788,8 @@ enter_notify (XEvent *ev)
 	if (w != 0 && w->mapped && w->visible
 	    && ev->xcrossing.detail != NotifyInferior)
 	{
-	    Fcall_window_hook (Qenter_notify_hook, rep_VAL(w), Qnil, Qnil);
+	    Fcall_window_hook (Qenter_notify_hook, rep_VAL(w),
+			       rep_LIST_1 (mode), Qnil);
 	}
     }
 }
@@ -725,8 +798,11 @@ static void
 leave_notify (XEvent *ev)
 {
     struct frame_part *fp;
+    repv mode = mode_to_sym (ev->xcrossing.mode);
     if (ev->xcrossing.window == root_window)
-	Fcall_hook (Qleave_notify_hook, Fcons (Qroot, Qnil), Qnil);
+    {
+	Fcall_hook (Qleave_notify_hook, rep_LIST_2 (Qroot, mode), Qnil);
+    }
     else if ((fp = find_frame_part_by_window (ev->xcrossing.window)) != 0)
     {
 	repv tem;
@@ -749,7 +825,7 @@ leave_notify (XEvent *ev)
 	if (tem && rep_CONSP(tem) && w->id != 0)
 	{
 	    Fcall_window_hook (Qleave_frame_part_hook, rep_VAL(w),
-			       Fcons (rep_VAL(fp), Qnil), Qnil);
+			       rep_LIST_2 (rep_VAL(fp), mode), Qnil);
 	}
     }
     else
@@ -758,7 +834,8 @@ leave_notify (XEvent *ev)
 	if (w != 0 && w->mapped && w->visible
 	    && ev->xcrossing.detail != NotifyInferior)
 	{
-	    Fcall_window_hook (Qleave_notify_hook, rep_VAL(w), Qnil, Qnil);
+	    Fcall_window_hook (Qleave_notify_hook, rep_VAL(w),
+			       rep_LIST_1 (mode), Qnil);
 	}
     }
 }
@@ -785,7 +862,11 @@ focus_in (XEvent *ev)
 		w->focus_change (w);
 	    }
 	    if (w->id != 0)
-		Fcall_window_hook (Qfocus_in_hook, rep_VAL(w), Qnil, Qnil);
+	    {
+		Fcall_window_hook (Qfocus_in_hook, rep_VAL(w),
+				   rep_LIST_1 (mode_to_sym (ev->xfocus.mode)),
+				   Qnil);
+	    }
 	}
     }
     else if (ev->xfocus.window == root_window)
@@ -812,7 +893,11 @@ focus_out (XEvent *ev)
 		w->focus_change (w);
 	    }
 	    if (w->id != 0)
-		Fcall_window_hook (Qfocus_out_hook, rep_VAL(w), Qnil, Qnil);
+	    {
+		Fcall_window_hook (Qfocus_out_hook, rep_VAL(w),
+				   rep_LIST_1 (mode_to_sym (ev->xfocus.mode)),
+				   Qnil);
+	    }
 	}
     }
 }
@@ -825,19 +910,15 @@ configure_request (XEvent *ev)
     {
 	XWindowChanges xwc;
 	u_int xwcm = (ev->xconfigurerequest.value_mask & 
-		      (CWX | CWY | CWWidth | CWHeight));
+		      (CWX | CWY | CWWidth | CWHeight
+		       | CWStackMode | CWSibling));
 	xwc.x = ev->xconfigurerequest.x;
 	xwc.y = ev->xconfigurerequest.y;
 	xwc.width = ev->xconfigurerequest.width;
 	xwc.height = ev->xconfigurerequest.height;
+	xwc.sibling = ev->xconfigurerequest.above;
+	xwc.stack_mode = ev->xconfigurerequest.detail;
 	XConfigureWindow (dpy, ev->xconfigurerequest.window, xwcm, &xwc);
-	if (ev->xconfigurerequest.value_mask & CWStackMode)
-	{
-	    if (ev->xconfigurerequest.detail == Above)
-		XRaiseWindow (dpy, ev->xconfigurerequest.window);
-	    else if (ev->xconfigurerequest.detail == Below)
-		XLowerWindow (dpy, ev->xconfigurerequest.window);
-	}
     }
     else if (w != 0)
     {
@@ -855,26 +936,50 @@ configure_request (XEvent *ev)
 	}
 	if (mask & CWStackMode)
 	{
-	    if (ev->xconfigurerequest.detail == Above)
-		alist = Fcons (Fcons (Qstack, Qabove), alist);
-	    else if (ev->xconfigurerequest.detail == Below)
-		alist = Fcons (Fcons (Qstack, Qbelow), alist);
-	}
-	if ((mask & CWX) || (mask & CWY))
-	{
-	    int x = w->attr.x, y = w->attr.y;
-	    if (mask & CWX)
-		x = (ev->xconfigurerequest.x
-		     + (w->reparented ? w->frame_x : 0));
-	    if (ev->xconfigurerequest.value_mask & CWY)
-		y = (ev->xconfigurerequest.y
-		     + (w->reparented ? w->frame_y : 0));
-	    if (x != w->attr.x || y != w->attr.y)
+	    Window above_win;
+	    repv relation = Qnil, sibling = Qnil;
+
+	    above_win = ev->xconfigurerequest.above;
+	    if (above_win != 0)
 	    {
-		alist = Fcons (Fcons (Qposition,
-				      Fcons (rep_MAKE_INT (x),
-					     rep_MAKE_INT (y))), alist);
+		Lisp_Window *tem = find_window_by_id (above_win);
+		if (tem != 0)
+		    sibling = rep_VAL (tem);
 	    }
+
+	    switch (ev->xconfigurerequest.detail)
+	    {
+	    case Above:
+		relation = Qabove;
+		break;
+	    case Below:
+		relation = Qbelow;
+		break;
+	    case TopIf:
+		relation = Qtop_if;
+		break;
+	    case BottomIf:
+		relation = Qbottom_if;
+		break;
+	    case Opposite:
+		relation = Qopposite;
+		break;
+	    }
+
+	    if (relation != Qnil)
+	    {
+		repv stack_data;
+		stack_data = Fcons (relation, Fcons (sibling, Qnil));
+		alist = Fcons (Fcons (Qstack, stack_data), alist);
+	    }
+	}
+	if ((mask & CWX) && (mask & CWY))
+	{
+	    int x = ev->xconfigurerequest.x;
+	    int y = ev->xconfigurerequest.y;
+	    alist = Fcons (Fcons (Qposition,
+				  Fcons (rep_MAKE_INT (x),
+					 rep_MAKE_INT (y))), alist);
 	}
 	if ((mask & CWWidth) || (mask & CWHeight))
 	{
@@ -892,8 +997,17 @@ configure_request (XEvent *ev)
 	}
 	if (alist != Qnil)
 	{
+	    /* Be sure to send one (and only one) synthetic ConfigureNotify
+	       to the window. The ICCCM states that we should send the event
+	       even if the state of the window doesn't change */
+
+	    Fsynthetic_configure_mutex (Qt);
+	    send_synthetic_configure (w);
+
 	    Fcall_window_hook (Qconfigure_request_hook, rep_VAL(w),
 			       rep_LIST_1 (alist), Qnil);
+
+	    Fsynthetic_configure_mutex (Qnil);
 	}
     }
 }
@@ -932,7 +1046,7 @@ shape_notify (XEvent *ev)
     if (w != 0 && sev->window == w->id && sev->kind == ShapeBounding)
     {
 	w->shaped = sev->shaped ? 1 : 0;
-	set_frame_shapes (w, TRUE);
+	queue_reshape_frame (w);
 	Fcall_window_hook (Qshape_notify_hook, rep_VAL(w), Qnil, Qnil);
     }
 }
@@ -973,9 +1087,9 @@ send_synthetic_configure (Lisp_Window *w)
 	w->pending_configure = 1;
 }
 
-DEFUN("synthetic-configure-mutex", Vsynthetic_configure_mutex,
+DEFUN("synthetic-configure-mutex", Fsynthetic_configure_mutex,
       Ssynthetic_configure_mutex, (repv arg), rep_Subr1) /*
-::doc:synthetic-configure-mutex::
+::doc:sawfish.wm.events#synthetic-configure-mutex::
 While this variable is non-nil no synthetic ConfigureNotify events will
 be sent to windows.
 ::end:: */
@@ -993,7 +1107,7 @@ be sent to windows.
     }
     return ret;
 }
-	
+
 long
 get_event_mask (int type)
 {
@@ -1119,14 +1233,8 @@ handle_input_mask(long mask)
 	current_event_window = rep_NULL;
 
 	rep_PUSHGC(gc_old_current_window, old_current_window);
-
-#if rep_INTERFACE >= 8
 	rep_call_with_barrier (inner_handle_input, rep_VAL (&xev),
 			       rep_TRUE, 0, 0, 0);
-#else
-	inner_handle_input (rep_VAL (&xev));
-#endif
-
 	rep_POPGC;
 
 	current_x_event = old_current_event;
@@ -1148,7 +1256,7 @@ handle_sync_input (int fd)
 /* Lisp functions */
 
 DEFUN("query-pointer", Fquery_pointer, Squery_pointer, (repv get), rep_Subr1) /*
-::doc:query-pointer::
+::doc:sawfish.wm.events#query-pointer::
 query-pointer [GET-FROM-SERVER]
 
 Returns (MOUSE-X . MOUSE-Y) representing the current mouse pointer position,
@@ -1176,7 +1284,7 @@ the server, otherwise it's taken from the current event (if possible).
 
 DEFUN("query-button-press-pointer", Fquery_button_press_pointer,
       Squery_button_press_pointer, (void), rep_Subr0) /*
-::doc:query-button-press-pointer::
+::doc:sawfish.wm.events#query-button-press-pointer::
 query-button-press-pointer
 
 Returns (MOUSE-X . MOUSE-Y) representing the mouse position relative to
@@ -1194,7 +1302,7 @@ the root window at the last button-press event.
 
 DEFUN("query-button-press-window", Fquery_button_press_window,
       Squery_button_press_window, (void), rep_Subr0) /*
-::doc:query-button-press-window::
+::doc:sawfish.wm.events#query-button-press-window::
 query-button-press-window
 
 Returns the window that the mouse was in when the button was pressed.
@@ -1216,7 +1324,7 @@ Returns the window that the mouse was in when the button was pressed.
 }
 
 DEFUN("query-pointer-window", Fquery_pointer_window, Squery_pointer_window, (void), rep_Subr0) /*
-::doc:query-pointer-window::
+::doc:sawfish.wm.events#query-pointer-window::
 query-pointer-window
 
 Returns the top-level window under the mouse pointer, or nil if the cursor
@@ -1245,7 +1353,7 @@ is in the root window.
 }
 
 DEFUN("accept-x-input", Faccept_x_input, Saccept_x_input, (repv mask), rep_Subr1) /*
-::doc:accept-x-input::
+::doc:sawfish.wm.events#accept-x-input::
 accept-x-input [EVENT-MASK]
 
 Handle any X events received. If EVENT-MASK is non-nil then only events
@@ -1258,7 +1366,7 @@ matching this numeric value are handled (see <X11/X.h>).
 
 DEFUN("current-event-window", Fcurrent_event_window, Scurrent_event_window,
       (repv win), rep_Subr1) /*
-::doc:current-event-window::
+::doc:sawfish.wm.events#current-event-window::
 current-event-window
 
 Return the window that received the current event, or the symbol
@@ -1295,11 +1403,10 @@ Return the window that received the current event, or the symbol
 
 DEFUN("x-server-timestamp", Fx_server_timestamp, Sx_server_timestamp,
       (repv from_server), rep_Subr1) /*
-::doc:x-server-timestamp::
+::doc:sawfish.wm.events#x-server-timestamp::
 x-server-timestamp [FROM-SERVER]
 
-Return a recent X server timestamp, as a cons cell containing two
-integers.
+Return a recent X server timestamp, as an integer.
 
 If FROM-SERVER is non-nil the timestamp is read directly from the
 server, otherwise the most recent timestamp seen by the window manager
@@ -1328,6 +1435,8 @@ DEFUN("clicked-frame-part", Fclicked_frame_part,
 void
 events_init (void)
 {
+    repv tem;
+
     event_handlers[VisibilityNotify] = visibility_notify;
     event_handlers[ColormapNotify] = colormap_notify;
     event_handlers[KeyPress] = key_press;
@@ -1424,6 +1533,7 @@ events_init (void)
     event_masks[ClientMessage] = 0;
     event_masks[MappingNotify] = StructureNotifyMask | SubstructureNotifyMask;
 
+    tem = rep_push_structure ("sawfish.wm.events");
     rep_ADD_SUBR(Squery_pointer);
     rep_ADD_SUBR(Squery_button_press_pointer);
     rep_ADD_SUBR(Squery_button_press_window);
@@ -1434,6 +1544,7 @@ events_init (void)
     rep_ADD_SUBR(Sx_events_queued);
     rep_ADD_SUBR(Sclicked_frame_part);
     rep_ADD_SUBR(Ssynthetic_configure_mutex);
+    rep_pop_structure (tem);
 
     rep_INTERN_SPECIAL(visibility_notify_hook);
     rep_INTERN_SPECIAL(destroy_notify_hook);
@@ -1452,20 +1563,33 @@ events_init (void)
     rep_INTERN_SPECIAL(enter_frame_part_hook);
     rep_INTERN_SPECIAL(leave_frame_part_hook);
     rep_INTERN_SPECIAL(configure_request_hook);
+    rep_INTERN_SPECIAL(window_state_change_hook);
+
+    rep_INTERN_SPECIAL(pointer_motion_threshold);
+    Fset (Qpointer_motion_threshold, rep_MAKE_INT (0));
 
     rep_INTERN(iconify_window);
     rep_INTERN(uniconify_window);
     rep_INTERN(root);
     rep_INTERN(new_value);
     rep_INTERN(deleted);
+    rep_INTERN(urgency);
+
     rep_INTERN(stack);
     rep_INTERN(above);
     rep_INTERN(below);
+    rep_INTERN(bottom_if);
+    rep_INTERN(top_if);
+    rep_INTERN(opposite);
     rep_INTERN(dimensions);
 
+    rep_INTERN(normal);
+    rep_INTERN(grab);
+    rep_INTERN(ungrab);
+
+    rep_INTERN(ignore_fp_keymap);
+
     rep_mark_static (&current_event_window);
-    current_context_map = Qnil;
-    rep_mark_static (&current_context_map);
 
     event_handler_context = XUniqueContext ();
 
