@@ -1,5 +1,5 @@
 /* frames.c -- window frame manipulation
-   $Id: frames.c,v 1.85 2000/06/07 09:35:04 john Exp $
+   $Id: frames.c,v 1.99 2000/11/07 20:15:06 jsh Exp $
 
    Copyright (C) 1999, 2000 John Harper <john@dcs.warwick.ac.uk>
 
@@ -83,6 +83,8 @@ DEFSYM(removed_classes, "removed-classes");
 DEFSYM(frame_part_classes, "frame-part-classes");
 DEFSYM(override_frame_part_classes, "override-frame-part-classes");
 DEFSYM(below_client, "below-client");
+DEFSYM(scale_foreground, "scale-foreground");
+DEFSYM(hidden, "hidden");
 
 static repv state_syms[fps_MAX];
 
@@ -218,6 +220,7 @@ fp_sweep (void)
 	cursor . CURSOR-OR-CURSOR-DEF
 
 	below-client . t
+	scale-foreground . t
 
    STATE's are one of: inactive, focused, highlighted, clicked,
    inactive-highlighted, inactive-clicked.
@@ -266,7 +269,9 @@ call_protectedly (repv (*fun)(repv), repv arg, repv error_return)
         if (stream != rep_NULL)
         {
 	    rep_stream_puts (stream, "frame error: ", -1, rep_FALSE);
-            rep_print_val (stream, rep_CDR(throw));
+	    if (rep_CAR (throw) == Qerror)
+		throw = rep_CDR (throw);
+            rep_print_val (stream, throw);
             rep_stream_putc (stream, '\n');
         }
 	result = error_return;
@@ -282,6 +287,33 @@ call_protectedly_1 (repv fun, repv arg, repv error_return)
     return call_protectedly (Ffuncall, rep_LIST_2 (fun, arg), error_return);
 }
 
+static void
+apply_mask (Drawable dest, int x_off, int y_off,
+	    int dest_width, int dest_height,
+	    Pixmap src, int src_width, int src_height)
+{
+    if (dest_width - x_off >= src_width
+	&& dest_height - y_off >= src_height)
+    {
+	/* No need for a temporary buffer */
+	XShapeCombineMask (dpy, dest, ShapeBounding,
+			   x_off, y_off, src, ShapeUnion);
+    }
+    else
+    {
+	Pixmap tem = XCreatePixmap (dpy, src, dest_width - x_off,
+				    dest_height - y_off, 1);
+	XGCValues gcv;
+	GC gc = XCreateGC (dpy, tem, 0, &gcv);
+	XCopyArea (dpy, src, tem, gc, 0, 0,
+		   dest_width - x_off, dest_height - y_off, 0, 0);
+	XShapeCombineMask (dpy, dest, ShapeBounding,
+			   x_off, y_off, tem, ShapeUnion);
+	XFreeGC (dpy, gc);
+	XFreePixmap (dpy, tem);
+    }
+}
+
 /* Construct the frame window's shape mask from the union of all
    individual shapes (frame parts and the client window, if appropriate).
    If ATOMIC is true, then the frame shape is changed _once_ only, using
@@ -297,10 +329,15 @@ set_frame_shapes (Lisp_Window *w, bool atomic)
 
     if (atomic)
     {
-	shape_win = XCreateSimpleWindow (dpy, root_window, -100, -100,
-					 w->frame_width, w->frame_height,
-					 0, BlackPixel (dpy, screen_num),
-					 BlackPixel (dpy, screen_num));
+	XSetWindowAttributes wa;
+	int wamask;
+	wa.colormap = image_cmap;
+	wa.border_pixel = BlackPixel (dpy, screen_num);
+	wamask = CWColormap | CWBorderPixel;
+	shape_win = XCreateWindow (dpy, root_window, -100, -100,
+				   w->frame_width, w->frame_height,
+				   0, image_depth, InputOutput,
+				   image_visual, wamask, &wa);
     }
     else
 	shape_win = w->frame;
@@ -364,22 +401,25 @@ set_frame_shapes (Lisp_Window *w, bool atomic)
 
 	    if (mask != 0)
 	    {
-		XShapeCombineMask (dpy, shape_win, ShapeBounding,
-				   fp->x - w->frame_x, fp->y - w->frame_y,
-				   mask, ShapeUnion);
+		int width = !tiled ? fp->width : image_width (image);
+		int height = !tiled ? fp->height : image_height (image);
+		int x_off = fp->x - w->frame_x;
+		int y_off = fp->y - w->frame_y;
+
+		apply_mask (shape_win, x_off, y_off,
+			    x_off + fp->width, y_off + fp->height,
+			    mask, width, height);
+
 		if (tiled)
 		{
-		    int width = image_width (image);
-		    int height = image_height (image);
 		    int x = width, y = 0;
 		    while (y < fp->height)
 		    {
 			while (x < fp->width)
 			{
-			    XShapeCombineMask (dpy, shape_win, ShapeBounding,
-					       fp->x - w->frame_x + x,
-					       fp->y - w->frame_y + y,
-					       mask, ShapeUnion);
+			    apply_mask (shape_win, x_off + x, y_off + y,
+					x_off + fp->width, y_off + fp->height,
+					mask, width, height);
 			    x += width;
 			}
 			y += height;
@@ -419,6 +459,27 @@ set_frame_shapes (Lisp_Window *w, bool atomic)
 			    0, 0, shape_win, ShapeBounding, ShapeSet);
 	XDestroyWindow (dpy, shape_win);
     }
+
+    w->pending_reshape = 0;
+}
+
+/* Queue an atomic reshape for the frame of W. It will happen sometime
+   in the near future. */
+void
+queue_reshape_frame (Lisp_Window *w)
+{
+    w->pending_reshape = 1;
+}
+
+void
+commit_queued_reshapes (void)
+{
+    Lisp_Window *w;
+    for (w = window_list; w != 0; w = w->next)
+    {
+	if (w->id != 0 && w->pending_reshape)
+	    set_frame_shapes (w, TRUE);
+    }
 }
 
 /* Draw the background of the frame-part FP. This is either a solid color
@@ -483,7 +544,6 @@ set_frame_part_bg (struct frame_part *fp)
 	    image_render (image, fp->width, fp->height, &bg_pixmap, &bg_mask);
 	}
 
-
 	/* Some of the Imlib_ functions call XSync on our display. In turn
 	   this can cause the error handler to run if a window has been
 	   deleted. This then invalidates the window we're updating */
@@ -507,11 +567,15 @@ set_frame_part_bg (struct frame_part *fp)
 	    if (bg_mask != 0)
 	    {
 		XRectangle rect;
-		tem = XCreateSimpleWindow (dpy, win->frame,
-					   -100, -100,
-					   fp->width, fp->height,
-					   0, BlackPixel (dpy, screen_num),
-					   BlackPixel (dpy, screen_num));
+		XSetWindowAttributes wa;
+		int wamask;
+		wa.colormap = image_cmap;
+		wa.border_pixel = BlackPixel (dpy, screen_num);
+		wamask = CWColormap | CWBorderPixel;
+		tem = XCreateWindow (dpy, win->frame, -100, -100,
+				     fp->width, fp->height,
+				     0, image_depth, InputOutput,
+				     image_visual, wamask, &wa);
 		rect.x = rect.y = 0;
 		rect.width = fp->width;
 		rect.height = fp->height;
@@ -551,6 +615,8 @@ set_frame_part_bg (struct frame_part *fp)
 
 	fp->drawn.bg = bg;
 	fp->drawn.fg = rep_NULL;
+
+	queue_reshape_frame (fp->win);
     }
     else if (Ffunctionp (bg) != Qnil)
     {
@@ -584,18 +650,26 @@ set_frame_part_fg (struct frame_part *fp)
     else if (IMAGEP(fg) || fp->text != Qnil)
     {
 	if (!COLORP(fg) && !IMAGEP(fg))
-	    fg = Fsymbol_value (Qdefault_foreground, Qt);
+	    fg = global_symbol_value (Qdefault_foreground);
 	if (!FONTP(font))
 	{
-	    font = Fsymbol_value (Qdefault_font, Qt);
+	    font = global_symbol_value (Qdefault_font);
 	    if (!FONTP(font))
 		goto out;
 	}
 
 	if (IMAGEP(fg))
 	{
-	    width = image_width (VIMAGE(fg));
-	    height = image_height (VIMAGE(fg));
+	    if (fp->scale_foreground)
+	    {
+		width = fp->width;
+		height = fp->height;
+	    }
+	    else
+	    {
+		width = image_width (VIMAGE(fg));
+		height = image_height (VIMAGE(fg));
+	    }
 	}
 	else
 	{
@@ -661,8 +735,7 @@ set_frame_part_fg (struct frame_part *fp)
 		set_frame_part_bg (fp);
 	    }
 
-	    image_render (VIMAGE(fg),
-			  image_width (VIMAGE(fg)), image_height (VIMAGE(fg)),
+	    image_render (VIMAGE(fg), width, height,
 			  &fg_pixmap, &fg_mask);
 
 	    /* Some of the Imlib_ functions call XSync on our display. In turn
@@ -673,9 +746,6 @@ set_frame_part_fg (struct frame_part *fp)
 
 	    if (fg_pixmap)
 	    {
-		int width = image_width (VIMAGE(fg));
-		int height = image_height (VIMAGE(fg));
-
 		if (fg_mask)
 		{
 		    gcv.clip_mask = fg_mask;
@@ -878,7 +948,7 @@ x_fp_assq (struct frame_part *fp, repv prop)
     else
 	class = rep_NULL;
 
-    tem = Fsymbol_value (Qoverride_frame_part_classes, Qt);
+    tem = global_symbol_value (Qoverride_frame_part_classes);
     if (class != rep_NULL && rep_CONSP(tem))
     {
 	tem = Fassq (class, tem);
@@ -891,7 +961,7 @@ x_fp_assq (struct frame_part *fp, repv prop)
 
     if (ret && ret == Qnil)
     {
-	tem = Fsymbol_value (Qframe_part_classes, Qt);
+	tem = global_symbol_value (Qframe_part_classes);
 	if (class != rep_NULL && rep_CONSP(tem))
 	{
 	    tem = Fassq (class, tem);
@@ -913,7 +983,23 @@ get_integer_prop (struct frame_part *fp, repv prop, repv class, repv ov_class)
 	    tem = rep_CDR(tem);
 	else
 	    tem = call_protectedly_1 (rep_CDR(tem), rep_VAL(fp->win), Qnil);
-	return (tem && rep_INTP(tem)) ? tem : Qnil;
+	return rep_INTP(tem) ? tem : Qnil;
+    }
+    else
+	return Qnil;
+}
+
+static repv
+get_boolean_prop (struct frame_part *fp, repv prop, repv class, repv ov_class)
+{
+    repv tem = fp_assq (fp, prop, class, ov_class);
+    if (tem && tem != Qnil)
+    {
+	if (Ffunctionp (rep_CDR(tem)) == Qnil)
+	    tem = rep_CDR(tem);
+	else
+	    tem = call_protectedly_1 (rep_CDR(tem), rep_VAL(fp->win), Qnil);
+	return tem;
     }
     else
 	return Qnil;
@@ -933,8 +1019,6 @@ get_pattern_prop (struct frame_part *fp, repv *data, repv (*conv)(repv data),
 	    tem = call_protectedly_1 (rep_CDR(tem), rep_VAL(fp->win), Qnil);
 	else
 	    tem = rep_CDR(tem);
-	if (!tem)
-	    return FALSE;
 	if (!rep_CONSP(tem))
 	{
 	    /* single value pattern */
@@ -989,12 +1073,7 @@ get_pattern_prop (struct frame_part *fp, repv *data, repv (*conv)(repv data),
 	for (i = 0; i < fps_MAX; i++)
 	{
 	    if (rep_STRINGP(data[i]))
-	    {
-		repv tem = call_protectedly (conv, data[i], Qnil);
-		if (tem == rep_NULL)
-		    return FALSE;
-		data[i] = tem;
-	    }
+		data[i] = call_protectedly (conv, data[i], Qnil);
 	}
 
 	/* now fill any gaps in the state table */
@@ -1036,14 +1115,14 @@ build_frame_part (struct frame_part *fp)
     if (tem && tem != Qnil)
     {
 	class = rep_CDR(tem);
-	tem = Fsymbol_value (Qframe_part_classes, Qt);
+	tem = global_symbol_value (Qframe_part_classes);
 	if (rep_CONSP(tem))
 	{
 	    tem = Fassq (class, tem);
 	    if (tem && tem != Qnil)
 		class_elt = rep_CDR(tem);
 	}
-	tem = Fsymbol_value (Qoverride_frame_part_classes, Qt);
+	tem = global_symbol_value (Qoverride_frame_part_classes);
 	if (rep_CONSP(tem))
 	{
 	    tem = Fassq (class, tem);
@@ -1053,6 +1132,9 @@ build_frame_part (struct frame_part *fp)
     }
 
     /* do we ignore this part? */
+    tem = get_boolean_prop (fp, Qhidden, class_elt, ov_class_elt);
+    if (tem != Qnil)
+	goto next_part;
     tem = Fassq (Qremovable, fp->alist);
     if (tem && tem != Qnil && rep_CDR(tem) != Qnil)
     {
@@ -1065,8 +1147,10 @@ build_frame_part (struct frame_part *fp)
 	}
     }
 
-    tem = fp_assq (fp, Qbelow_client, class_elt, ov_class_elt);
+    tem = get_boolean_prop (fp, Qbelow_client, class_elt, ov_class_elt);
     fp->below_client = (tem && tem != Qnil);
+    tem = get_boolean_prop (fp, Qscale_foreground, class_elt, ov_class_elt);
+    fp->scale_foreground = (tem && tem != Qnil);
 
     /* get text label */
     tem = fp_assq (fp, Qtext, class_elt, ov_class_elt);
@@ -1080,8 +1164,7 @@ build_frame_part (struct frame_part *fp)
 	tem = rep_CDR(tem);
 	if (Ffunctionp (tem) != Qnil)
 	    tem = call_protectedly_1 (tem, rep_VAL(w), Qnil);
-	if (tem != rep_NULL)
-	    fp->x_justify = tem;
+	fp->x_justify = tem;
     }
     else
 	fp->x_justify = Qnil;
@@ -1091,8 +1174,7 @@ build_frame_part (struct frame_part *fp)
 	tem = rep_CDR(tem);
 	if (Ffunctionp (tem) != Qnil)
 	    tem = call_protectedly_1 (tem, rep_VAL(w), Qnil);
-	if (tem != rep_NULL)
-	    fp->y_justify = tem;
+	fp->y_justify = tem;
     }
     else
 	fp->y_justify = Qnil;
@@ -1103,9 +1185,11 @@ build_frame_part (struct frame_part *fp)
     if (tem != Qnil)
     {
 	tem = rep_CDR(tem);
+	if (Ffunctionp (tem) != Qnil)
+	    tem = call_protectedly_1 (tem, rep_VAL (w), Qnil);
 	if (!CURSORP(tem) && tem != Qnil)
 	    tem = call_protectedly (Fget_cursor, tem, Qnil);
-	if (tem && CURSORP(tem))
+	if (CURSORP(tem))
 	    fp->cursor = tem;
     }
 
@@ -1262,7 +1346,8 @@ configure_frame_part (struct frame_part *fp)
 	    wa.win_gravity = StaticGravity;
 	    wa.bit_gravity = StaticGravity;
 	    wa.colormap = image_cmap;
-	    wamask = CWWinGravity | CWBitGravity | CWColormap;
+	    wa.border_pixel = BlackPixel (dpy, screen_num);
+	    wamask = CWWinGravity | CWBitGravity | CWColormap | CWBorderPixel;
 	    fp->id = XCreateWindow (dpy, w->frame,
 				    fp->x - w->frame_x, fp->y - w->frame_y,
 				    fp->width, fp->height,
@@ -1330,6 +1415,7 @@ list_frame_generator (Lisp_Window *w)
     int nparts = 0;
     XSetWindowAttributes wa;
     u_long wamask;
+    int old_x_off, old_y_off;
 
     /* bounding box of frame */
     int left_x, top_y, right_x, bottom_y;
@@ -1419,6 +1505,9 @@ list_frame_generator (Lisp_Window *w)
     bigger = (right_x - left_x > w->frame_width
 	      || bottom_y - top_y > w->frame_height);
 
+    old_x_off = w->frame_x;
+    old_y_off = w->frame_y;
+
     /* now we can find the size and offset of the frame. */
     w->frame_width = right_x - left_x;
     w->frame_height = bottom_y - top_y;
@@ -1426,7 +1515,7 @@ list_frame_generator (Lisp_Window *w)
     w->frame_y = top_y;
 
     DB(("  bounding box: x=%d y=%d width=%d height=%d\n",
-	w->frame_x, w->frame_y, w->frame_width, w->frame_height));
+	left_x, top_y, w->frame_width, w->frame_height));
 
     if (w->reparented && bigger)
 	set_frame_shapes (w, TRUE);
@@ -1437,19 +1526,25 @@ list_frame_generator (Lisp_Window *w)
     {
 	/* create the frame */
 	wa.override_redirect = True;
-	wamask = CWOverrideRedirect;
+	wa.colormap = image_cmap;
+	wa.border_pixel = BlackPixel (dpy, screen_num);
+	wamask = CWOverrideRedirect | CWColormap | CWBorderPixel;
 	w->frame = XCreateWindow (dpy, root_window, w->attr.x, w->attr.y,
 				  w->frame_width, w->frame_height,
-				  0, screen_depth, InputOutput,
-				  screen_visual, wamask, &wa);
+				  0, image_depth, InputOutput,
+				  image_visual, wamask, &wa);
     }
     else
     {
+	/* adjust frame position to keep absolute client position constant */
+	w->attr.x += w->frame_x - old_x_off;
+	w->attr.y += w->frame_y - old_y_off;
+
 	XMoveResizeWindow (dpy, w->frame, w->attr.x, w->attr.y,
 			   w->frame_width, w->frame_height);
 
 	if (w->reparented)
-	    XMoveResizeWindow (dpy, w->id, -w->frame_x, -w->frame_y,
+	    XMoveResizeWindow (dpy, w->id, -left_x, -top_y,
 			       w->attr.width, w->attr.height);
 	else
 	    XResizeWindow (dpy, w->id, w->attr.width, w->attr.height);
@@ -1460,9 +1555,6 @@ list_frame_generator (Lisp_Window *w)
     w->rebuild_frame = list_frame_generator;
     w->property_change = refresh_frame_parts;
 
-    /* Client window is always left _underneath_ any overlapping frame
-       parts; this may not always be ideal, but we have to choose 
-       either over or under, and this will probably be more useful.. */
     if (w->reparented)
 	XLowerWindow (dpy, w->id);
 
@@ -1553,11 +1645,16 @@ void
 create_window_frame (Lisp_Window *w)
 {
     DB(("create_window_frame (%s)\n", rep_STR(w->name)));
-    w->destroy_frame = 0;
-    w->focus_change = 0;
-    w->rebuild_frame = 0;
-    w->property_change = 0;
-    list_frame_generator (w);
+    if (w->frame_parts == 0)
+    {
+	w->destroy_frame = 0;
+	w->focus_change = 0;
+	w->rebuild_frame = 0;
+	w->property_change = 0;
+	list_frame_generator (w);
+    }
+    else
+	fprintf (stderr, "warning: reframing framed window: %lx\n", (long) w->id);
 }
 
 /* Destroy the frame of window W. If LEAVE-FRAME-WIN is non-zero, then
@@ -1568,7 +1665,10 @@ destroy_window_frame (Lisp_Window *w, bool leave_frame_win)
     if (w->frame != 0)
     {
 	if (w->destroy_frame != 0)
+	{
 	    w->destroy_frame (w);
+	    w->destroy_frame = 0;
+	}
 	if (!leave_frame_win && w->frame != 0)
 	{
 	    XDestroyWindow (dpy, w->frame);
@@ -1582,7 +1682,7 @@ destroy_window_frame (Lisp_Window *w, bool leave_frame_win)
 
 DEFUN("frame-draw-mutex", Fframe_draw_mutex,
       Sframe_draw_mutex, (repv arg), rep_Subr1) /*
-::doc:frame-draw-mutex::
+::doc:sawfish.wm.frames.subrs#frame-draw-mutex::
 While this variable is non-nil no frame parts will be redrawn. When it is
 set to nil any pending redraws will take place.
 ::end:: */
@@ -1607,7 +1707,7 @@ set to nil any pending redraws will take place.
 	
 DEFUN("frame-state-mutex", Fframe_state_mutex,
       Sframe_state_mutex, (repv arg), rep_Subr1) /*
-::doc:frame-state-mutex::
+::doc:sawfish.wm.frames.subrs#frame-state-mutex::
 While this variable is non-nil the state of frame parts will not be
 altered when the pointer enters or leaves its window.
 ::end:: */
@@ -1627,7 +1727,7 @@ altered when the pointer enters or leaves its window.
 
 DEFUN("frame-part-get", Fframe_part_get,
       Sframe_part_get, (repv part, repv prop), rep_Subr2) /*
-::doc:frame-part-get::
+::doc:sawfish.wm.frames.subrs#frame-part-get::
 frame-part-get PART PROPERTY
 ::end:: */
 {
@@ -1734,7 +1834,7 @@ DEFUN("rebuild-frame-part", Frebuild_frame_part,
     {
 	/* XXX what about reconfiguring the container window..? */
 	configure_frame_part (VPART(part));
-	set_frame_shapes (VPART(part)->win, TRUE);
+	queue_reshape_frame (VPART(part)->win);
     }
     rep_POPGC;
     return Qt;
@@ -1755,10 +1855,12 @@ DEFUN("refresh-window", Frefresh_window,
 void
 frames_init (void)
 {
+    repv tem;
     frame_part_type = rep_register_new_type ("frame-part", fp_cmp, fp_prin,
 					     fp_prin, fp_sweep, fp_mark, 0,
 					     0, 0, 0, 0, 0, 0);
 
+    tem = rep_push_structure ("sawfish.wm.frames.subrs");
     rep_ADD_SUBR(Sframe_draw_mutex);
     rep_ADD_SUBR(Sframe_state_mutex);
     rep_ADD_SUBR(Sframe_part_get);
@@ -1771,7 +1873,11 @@ frames_init (void)
     rep_ADD_SUBR(Smap_frame_parts);
     rep_ADD_SUBR(Srefresh_frame_part);
     rep_ADD_SUBR(Srebuild_frame_part);
+    rep_pop_structure (tem);
+
+    tem = rep_push_structure ("sawfish.wm.windows.subrs");
     rep_ADD_SUBR(Srefresh_window);
+    rep_pop_structure (tem);
 
     rep_INTERN(internal);
     rep_INTERN(tiled);
@@ -1806,6 +1912,8 @@ frames_init (void)
     rep_INTERN(removable);
     rep_INTERN(removed_classes);
     rep_INTERN(below_client);
+    rep_INTERN(scale_foreground);
+    rep_INTERN(hidden);
 
     rep_INTERN_SPECIAL(frame_part_classes);
     rep_INTERN_SPECIAL(override_frame_part_classes);
