@@ -1,5 +1,5 @@
 /* functions.c -- useful window manager Lisp functions
-   $Id: functions.c,v 1.74 2000/06/07 09:34:41 john Exp $
+   $Id: functions.c,v 1.83 2000/10/23 04:09:07 john Exp $
 
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
@@ -19,6 +19,9 @@
    along with sawmill; see the file COPYING.   If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
+/* Define this to fake two side-by-side Xinerama style heads */
+#undef TEST_XINERAMA
+
 /* AIX requires this to be the first thing in the file.  */
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -49,16 +52,25 @@ static int xinerama_heads;
 #ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
 # include <X11/extensions/Xinerama.h>
   static XineramaScreenInfo *xinerama_head_info;
+# ifdef TEST_XINERAMA
+   static XineramaScreenInfo debug_heads[2] = {
+     { 0, 0, 0, 512, 768 },
+     { 0, 512, 0, 512, 768 }
+   };
+   static int debug_nheads = 2;
+# endif
 #endif
 
 DEFSYM(root, "root");
 DEFSYM(after_restacking_hook, "after-restacking-hook");
 DEFSYM(position, "position");
 DEFSYM(spacing, "spacing");
+DEFSYM(window, "window");
+DEFSYM(head, "head");
 
 DEFUN("restack-windows", Frestack_windows, Srestack_windows,
       (repv list), rep_Subr1) /*
-::doc:restack-windows::
+::doc:sawfish.wm.windows.subrs#restack-windows::
 restack-windows LIST
 
 Restack all windows in the list of windows LIST in the order they occur
@@ -121,7 +133,7 @@ do_raise_lower (repv win, repv sib, u_int mode)
 
 DEFUN("x-raise-window", Fx_raise_window, Sx_raise_window,
       (repv win, repv above), rep_Subr2) /*
-::doc:x-raise-window::
+::doc:sawfish.wm.windows.subrs#x-raise-window::
 x-raise-window WINDOW [ABOVE]
 
 Raise WINDOW so that it is above window ABOVE. If ABOVE is undefined,
@@ -133,7 +145,7 @@ raise WINDOW to the top of the stacking order.
 
 DEFUN("x-lower-window", Fx_lower_window, Sx_lower_window,
       (repv win, repv below), rep_Subr2) /*
-::doc:x-lower-window::
+::doc:sawfish.wm.windows.subrs#x-lower-window::
 x-lower-window WINDOW [BELOW]
 
 Lower WINDOW so that it is below window BELOW. If BELOW is undefined,
@@ -154,7 +166,7 @@ DEFUN("x-kill-client", Fx_kill_client, Sx_kill_client,
 }
 
 DEFUN_INT("destroy-window", Fdestroy_window, Sdestroy_window, (repv win), rep_Subr1, "%W") /*
-::doc:destroy-window::
+::doc:sawfish.wm.windows.subrs#destroy-window::
 destroy-window WINDOW
 
 Destroy WINDOW with out giving the owning application any warning.
@@ -172,7 +184,7 @@ WINDOW may be a window object or a numeric window id.
 }
 
 DEFUN("warp-cursor", Fwarp_cursor, Swarp_cursor, (repv x, repv y), rep_Subr2) /*
-::doc:warp-cursor::
+::doc:sawfish.wm.misc#warp-cursor::
 warp-cursor X Y
 
 Move the mouse pointer to position (X, Y) relative to the origin of the
@@ -194,7 +206,7 @@ root window.
 
 DEFUN("move-window-to", Fmove_window_to, Smove_window_to,
       (repv win, repv x, repv y), rep_Subr3) /*
-::doc:move-window-to::
+::doc:sawfish.wm.windows.subrs#move-window-to::
 move-window-to WINDOW X Y
 
 Move the top-left corner of window object WINDOW to (X, Y).
@@ -218,7 +230,7 @@ Move the top-left corner of window object WINDOW to (X, Y).
 
 DEFUN("resize-window-to", Fresize_window_to, Sresize_window_to,
       (repv win, repv width, repv height), rep_Subr3) /*
-::doc:resize-window-to::
+::doc:sawfish.wm.windows.subrs#resize-window-to::
 resize-window-to WINDOW WIDTH HEIGHT
 
 Set the dimensions of window object WINDOW to (WIDTH, HEIGHT).
@@ -236,8 +248,8 @@ Set the dimensions of window object WINDOW to (WIDTH, HEIGHT).
 
 DEFUN("move-resize-window-to", Fmove_resize_window_to, Smove_resize_window_to,
       (repv win, repv x, repv y, repv width, repv height), rep_Subr5) /*
-::doc:move-resize-window-to::
-move-resize-window-to X Y WINDOW WIDTH HEIGHT
+::doc:sawfish.wm.windows.subrs#move-resize-window-to::
+move-resize-window-to WINDOW X Y WIDTH HEIGHT
 
 Reconfigure the geometry of window object WINDOW as specified.
 ::end:: */
@@ -272,7 +284,7 @@ Reconfigure the geometry of window object WINDOW as specified.
 }
 
 DEFUN("grab-server", Fgrab_server, Sgrab_server, (void), rep_Subr0) /*
-::doc:grab-server::
+::doc:sawfish.wm.misc#grab-server::
 grab-server
 
 Prevent any other clients from accessing the X server. See `ungrab-server'.
@@ -288,7 +300,7 @@ Prevent any other clients from accessing the X server. See `ungrab-server'.
 }
 
 DEFUN("ungrab-server", Fungrab_server, Sungrab_server, (void), rep_Subr0) /*
-::doc:ungrab-server::
+::doc:sawfish.wm.misc#ungrab-server::
 ungrab-server
 
 After a call to `grab-server' this will allow other clients to access
@@ -307,7 +319,7 @@ Note that calls to `grab-server' and `ungrab-server' _nest_.
 
 DEFUN("server-grabbed-p", Fserver_grabbed_p,
       Sserver_grabbed_p, (void), rep_Subr0) /*
-::doc:server-grabbed-p::
+::doc:sawfish.wm.misc#server-grabbed-p::
 server-grabbed-p
 
 Return t if the server is currently grabbed.
@@ -331,7 +343,7 @@ regrab_server (void)
 DEFUN("grab-pointer", Fgrab_pointer, Sgrab_pointer,
       (repv win, repv cursor, repv ptr_sync, repv kbd_sync, repv confine_to),
       rep_Subr5) /*
-::doc:grab-pointer::
+::doc:sawfish.wm.events#grab-pointer::
 grab-pointer [WINDOW] [CURSOR] [PTR-SYNC] [KBD-SYNC] [CONFINE-TO]
 
 Grab the pointer and direct all pointer events to window object
@@ -399,7 +411,7 @@ again:
 }
 
 DEFUN("ungrab-pointer", Fungrab_pointer, Sungrab_pointer, (void), rep_Subr0) /*
-::doc:ungrab-pointer::
+::doc:sawfish.wm.events#ungrab-pointer::
 ungrab-pointer
 
 Release the grab on the mouse pointer.
@@ -413,7 +425,7 @@ Release the grab on the mouse pointer.
 
 DEFUN("grab-keyboard", Fgrab_keyboard, Sgrab_keyboard,
       (repv win, repv ptr_sync, repv kbd_sync), rep_Subr3) /*
-::doc:grab-keyboard::
+::doc:sawfish.wm.events#grab-keyboard::
 grab-keyboard [WINDOW] [PTR-SYNC] [KBD-SYNC]
 
 Grab the keyboard and direct all keyboard events to window object
@@ -458,7 +470,7 @@ again:
     
 DEFUN("ungrab-keyboard", Fungrab_keyboard,
       Sungrab_keyboard, (void), rep_Subr0) /*
-::doc:ungrab-keyboard::
+::doc:sawfish.wm.events#ungrab-keyboard::
 ungrab-keyboard
 
 Release the grab on the keyboard.
@@ -470,7 +482,7 @@ Release the grab on the keyboard.
 }
 
 DEFUN("screen-width", Fscreen_width, Sscreen_width, (void), rep_Subr0) /*
-::doc:screen-width::
+::doc:sawfish.wm.misc#screen-width::
 screen-width
 
 Return the width of the root window (in pixels).
@@ -480,7 +492,7 @@ Return the width of the root window (in pixels).
 }
 
 DEFUN("screen-height", Fscreen_height, Sscreen_height, (void), rep_Subr0) /*
-::doc:screen-height::
+::doc:sawfish.wm.misc#screen-height::
 screen-height
 
 Return the height of the root window (in pixels).
@@ -490,7 +502,7 @@ Return the height of the root window (in pixels).
 }
 
 DEFUN("sync-server", Fsync_server, Ssync_server, (void), rep_Subr0) /*
-::doc:sync-server::
+::doc:sawfish.wm.misc#sync-server::
 sync-server
 
 Flush all pending X requests, don't wait for them to finish.
@@ -502,7 +514,7 @@ Flush all pending X requests, don't wait for them to finish.
 
 DEFUN("delete-x-property", Fdelete_x_property, Sdelete_x_property,
       (repv win, repv atom), rep_Subr2) /*
-::doc:delete-x-property::
+::doc:sawfish.wm.misc#delete-x-property::
 delete-x-property WINDOW ATOM
 
 Delete the X property ATOM (a symbol) of WINDOW.
@@ -521,7 +533,7 @@ WINDOW may be the symbol `root', a window object or a numeric window id.
 
 DEFUN("list-x-properties", Flist_x_properties, Slist_x_properties,
       (repv win), rep_Subr1) /*
-::doc:list-x-properties::
+::doc:sawfish.wm.misc#list-x-properties::
 list-x-properties WINDOW
 
 List all X properties (symbols) of WINDOW.
@@ -565,7 +577,7 @@ WINDOW may be the symbol `root', a window object or a numeric window id.
 
 DEFUN("get-x-property", Fget_x_property, Sget_x_property,
       (repv win, repv prop), rep_Subr2) /*
-::doc:get-x-property::
+::doc:sawfish.wm.misc#get-x-property::
 get-x-property WINDOW PROPERTY
 
 Return (TYPE FORMAT DATA) representing the X property PROPERTY (a
@@ -665,7 +677,7 @@ symbols, representing the atoms read.
 
 DEFUN("set-x-property", Fset_x_property, Sset_x_property,
       (repv win, repv prop, repv data, repv type, repv format), rep_Subr5) /*
-::doc:set-x-property::
+::doc:sawfish.wm.misc#set-x-property::
 set-x-property WINDOW PROPERTY DATA TYPE FORMAT
 
 Set the X property PROPERTY (a symbol) of WINDOW to the array DATA.
@@ -752,7 +764,7 @@ converted to their numeric X atoms.
 
 DEFUN("get-x-text-property", Fget_x_text_property, Sget_x_text_property,
       (repv win, repv prop), rep_Subr2) /*
-::doc:get-x-text-property::
+::doc:sawfish.wm.misc#get-x-text-property::
 get-x-text-property WINDOW PROPERTY
 ::end:: */
 {
@@ -787,7 +799,7 @@ get-x-text-property WINDOW PROPERTY
 
 DEFUN("set-x-text-property", Fset_x_text_property, Sset_x_text_property, 
       (repv win, repv prop, repv vect), rep_Subr3) /*
-::doc:set-x-text-prooperty::
+::doc:sawfish.wm.misc#set-x-text-prooperty::
 set-x-text-property WINDOW PROPERTY STRING-VECTOR
 ::end:: */
 {
@@ -823,7 +835,7 @@ set-x-text-property WINDOW PROPERTY STRING-VECTOR
 
 DEFUN("send-client-message", Fsend_client_message, Ssend_client_message,
       (repv win, repv type, repv data, repv format), rep_Subr4) /*
-::doc:send-client-message::
+::doc:sawfish.wm.events#send-client-message::
 send-client-message WINDOW TYPE DATA FORMAT
 
 Send an X ClientMessage event to WINDOW (a window object or the symbol
@@ -882,7 +894,7 @@ FORMAT sized quantities (8, 16 or 32).
 
 DEFUN("create-window", Fcreate_window, Screate_window,
       (repv parent, repv x, repv y, repv width, repv height), rep_Subr5) /*
-::doc:create-window::
+::doc:sawfish.wm.misc#create-window::
 create-window PARENT-WINDOW X Y WIDTH HEIGHT
 
 Create an unmapped window that is a child of PARENT-WINDOW (a window object,
@@ -907,7 +919,7 @@ Returns the window id of the new window.
 }
 
 DEFUN("x-atom", Fx_atom, Sx_atom, (repv symbol), rep_Subr1) /*
-::doc:x-atom::
+::doc:sawfish.wm.misc#x-atom::
 x-atom SYMBOL
 
 Return the integer identifying the X atom with the same name as SYMBOL.
@@ -919,7 +931,7 @@ Return the integer identifying the X atom with the same name as SYMBOL.
 }
 
 DEFUN("x-atom-name", Fx_atom_name, Sx_atom_name, (repv atom), rep_Subr1) /*
-::doc:x-atom-name::
+::doc:sawfish.wm.misc#x-atom-name::
 x-atom-name ATOM
 
 Return the symbol with the same name as the X atom identified by the
@@ -930,6 +942,16 @@ integer ATOM.
     return x_atom_symbol (rep_INT(atom));
 }
 
+DEFUN("root-window-id", Froot_window_id, Sroot_window_id, (void), rep_Subr0) /*
+::doc:sawfish.wm.misc#root-window-id::
+root-window-id
+
+Returns the numeric id of the root window of the managed screen.
+::end:: */
+{
+    return rep_MAKE_INT (root_window);
+}
+
 
 /* xinerama support */
 
@@ -952,12 +974,12 @@ DEFUN ("find-head", Ffind_head, Sfind_head, (repv x, repv y), rep_Subr2)
 #ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
     for (i = 0; i < xinerama_heads; i++)
     {
-	if ((xinerama_head_info[i].x_org >= rep_INT (x))
-	    && (xinerama_head_info[i].y_org >= rep_INT (y))
+	if ((xinerama_head_info[i].x_org <= rep_INT (x))
+	    && (xinerama_head_info[i].y_org <= rep_INT (y))
 	    && (xinerama_head_info[i].x_org
-		+ xinerama_head_info[i].width < rep_INT (y))
+		+ xinerama_head_info[i].width > rep_INT (x))
 	    && (xinerama_head_info[i].y_org
-		+ xinerama_head_info[i].height < rep_INT (y)))
+		+ xinerama_head_info[i].height > rep_INT (y)))
 	{
 	    return rep_MAKE_INT (i);
 	}
@@ -1139,7 +1161,7 @@ DEFUN("display-message", Fdisplay_message, Sdisplay_message,
 	    }
 	}
 	if (!FONTP(message.font))
-	    message.font = Fsymbol_value (Qdefault_font, Qt);
+	    message.font = global_symbol_value (Qdefault_font);
 	if (!FONTP(message.font))
 	    return rep_signal_arg_error (Qfont, 1);
 
@@ -1186,24 +1208,60 @@ DEFUN("display-message", Fdisplay_message, Sdisplay_message,
 	{
 	    char *ptr = rep_STR(text);
 	    int max_width = 0, rows = 0;
+	    repv head = Qnil;
+	    int head_width, head_height;
+	    int head_xoff, head_yoff;
 	    while (*ptr != 0)
 	    {
 		int text_width;
 		char *end = strchr (ptr, '\n');
 		if (end == 0)
-		    end = ptr + strlen (ptr);
+		  end = ptr + strlen (ptr);
 		text_width = x_text_width (message.font, ptr, end - ptr);
 		max_width = MAX(max_width, text_width);
 		rows++;
 		ptr = end;
 		if (*ptr == '\n')
-		    ptr++;
+		  ptr++;
 	    }
 	    message.width = max_width + MSG_PAD_X * 2;
 	    height = ((rep_INT(Ffont_height (message.font)) + message.spacing)
 		      * rows + MSG_PAD_Y * 2);
-	    x = (screen_width - message.width) / 2;
-	    y = (screen_height - height) / 2;
+
+	    /* Find the head to put the message on. */
+	    tem = Fassq (Qhead, attrs);
+	    if (tem && rep_CONSP (tem) && rep_INTP (rep_CDR (tem)))
+		head = rep_CDR (tem);
+	    if (!head || head == Qnil)
+		head = Ffind_head (Fquery_pointer (Qnil), Qnil);
+
+	    if (head && head != Qnil)
+	    {
+		/* We have a head to centre on. */
+
+		tem = Fhead_dimensions (head);
+		if (!tem)
+		    goto no_head;
+		head_width = rep_INT (rep_CAR (tem));
+		head_height = rep_INT (rep_CDR (tem));
+
+		tem = Fhead_offset (head);
+		if (!tem)
+		    goto no_head;
+		head_xoff = rep_INT (rep_CAR (tem));
+		head_yoff = rep_INT (rep_CDR (tem));
+	    }
+	    else
+	    {
+		/* No head, just centre on the screen. */
+	    no_head:
+		head_xoff = head_yoff = 0;
+		head_width = screen_width;
+		head_height = screen_height;
+	    }
+
+	    x = head_xoff + ((head_width - message.width) / 2);
+	    y = head_yoff + ((head_height - height) / 2);
 	}
 
 	tem = Fassq (Qposition, attrs);
@@ -1282,22 +1340,24 @@ DEFUN("display-message", Fdisplay_message, Sdisplay_message,
 void
 functions_init (void)
 {
+    repv tem;
+
+    tem = rep_push_structure ("sawfish.wm.windows.subrs");
     rep_ADD_SUBR(Srestack_windows);
     rep_ADD_SUBR(Sx_raise_window);
     rep_ADD_SUBR(Sx_lower_window);
-    rep_ADD_SUBR(Sx_kill_client);
     rep_ADD_SUBR_INT(Sdestroy_window);
-    rep_ADD_SUBR(Swarp_cursor);
     rep_ADD_SUBR(Smove_window_to);
     rep_ADD_SUBR(Sresize_window_to);
     rep_ADD_SUBR(Smove_resize_window_to);
+    rep_pop_structure (tem);
+
+    tem = rep_push_structure ("sawfish.wm.misc");
+    rep_ADD_SUBR(Sx_kill_client);
+    rep_ADD_SUBR(Swarp_cursor);
     rep_ADD_SUBR(Sgrab_server);
     rep_ADD_SUBR(Sungrab_server);
     rep_ADD_SUBR(Sserver_grabbed_p);
-    rep_ADD_SUBR(Sgrab_pointer);
-    rep_ADD_SUBR(Sungrab_pointer);
-    rep_ADD_SUBR(Sgrab_keyboard);
-    rep_ADD_SUBR(Sungrab_keyboard);
     rep_ADD_SUBR(Sscreen_width);
     rep_ADD_SUBR(Sscreen_height);
     rep_ADD_SUBR(Ssync_server);
@@ -1307,19 +1367,30 @@ functions_init (void)
     rep_ADD_SUBR(Sset_x_property);
     rep_ADD_SUBR(Sget_x_text_property);
     rep_ADD_SUBR(Sset_x_text_property);
-    rep_ADD_SUBR(Ssend_client_message);
     rep_ADD_SUBR(Screate_window);
     rep_ADD_SUBR(Sx_atom);
     rep_ADD_SUBR(Sx_atom_name);
+    rep_ADD_SUBR(Sroot_window_id);
     rep_ADD_SUBR(Shead_count);
     rep_ADD_SUBR(Sfind_head);
     rep_ADD_SUBR(Shead_dimensions);
     rep_ADD_SUBR(Shead_offset);
     rep_ADD_SUBR(Sdisplay_message);
+    rep_pop_structure (tem);
+
+    tem = rep_push_structure ("sawfish.wm.events");
+    rep_ADD_SUBR(Sgrab_pointer);
+    rep_ADD_SUBR(Sungrab_pointer);
+    rep_ADD_SUBR(Sgrab_keyboard);
+    rep_ADD_SUBR(Sungrab_keyboard);
+    rep_ADD_SUBR(Ssend_client_message);
+    rep_pop_structure (tem);
+
     rep_INTERN(root);
     rep_INTERN_SPECIAL(after_restacking_hook);
     rep_INTERN(position);
     rep_INTERN(spacing);
+    rep_INTERN(head);
 
     rep_mark_static (&message.text);
     rep_mark_static (&message.fg);
@@ -1328,8 +1399,13 @@ functions_init (void)
     rep_mark_static (&message.justify);
 
 #ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
+# ifndef TEST_XINERAMA
     if (dpy != 0)
 	xinerama_head_info = XineramaQueryScreens (dpy, &xinerama_heads);
+# else
+    xinerama_head_info = debug_heads;
+    xinerama_heads = debug_nheads;
+# endif
 #endif
 }
 
