@@ -1,5 +1,5 @@
 /* keys.c -- Key binding and evaluating (this should be called events.c)
-   $Id: keys.c,v 1.61 2000/06/10 14:50:09 john Exp $
+   $Id: keys.c,v 1.74 2000/11/13 23:57:43 jsh Exp $
 
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
@@ -23,6 +23,7 @@
 #include "keys.h"
 #include <string.h>
 #include <stdlib.h>
+#include <ctype.h>
 #include <X11/keysym.h>
 #include <X11/Xutil.h>
 
@@ -57,6 +58,11 @@ DEFSYM(override_keymap, "override-keymap");
 DEFSYM(unbound_key_hook, "unbound-key-hook");
 DEFSYM(keymap, "keymap");
 
+DEFSYM(this_command, "this-command");
+DEFSYM(last_command, "last-command");
+DEFSYM(prefix_arg, "prefix-arg");
+DEFSYM(current_prefix_arg, "current-prefix-arg");
+
 DEFSYM(async_pointer, "async-pointer");
 DEFSYM(async_keyboard, "async-keyboard");
 DEFSYM(sync_pointer, "sync-pointer");
@@ -67,6 +73,7 @@ DEFSYM(sync_both, "sync-both");
 DEFSYM(async_both, "async-both");
 
 DEFSYM(display_message, "display-message");
+DEFSYM(call_command, "call-command");
 
 static repv next_keymap_path;
 
@@ -106,7 +113,7 @@ translate_event(u_long *code, u_long *mods, XEvent *xev)
     switch(xev->type)
     {
     case KeyRelease:
-	if (Fsymbol_value (Qeval_key_release_events, Qt) == Qnil)
+	if (global_symbol_value (Qeval_key_release_events) == Qnil)
 	    break;
 	/* FALL THROUGH */
 
@@ -139,7 +146,7 @@ translate_event(u_long *code, u_long *mods, XEvent *xev)
 	    *code = XKeycodeToKeysym(xev->xany.display, xev->xkey.keycode, 0);
 	if(*code != NoSymbol
 	   && (!IsModifierKey (*code)
-	       || Fsymbol_value (Qeval_modifier_events, Qt) == Qt))
+	       || global_symbol_value (Qeval_modifier_events) == Qt))
 	{
 	    *mods |= EV_TYPE_KEY;
 	    ret = TRUE;
@@ -150,7 +157,7 @@ translate_event(u_long *code, u_long *mods, XEvent *xev)
 
 	if(xev->xbutton.button == last_click_button)
         {
-            multi_click_delay = Fsymbol_value (Qmulti_click_delay, Qt);
+            multi_click_delay = global_symbol_value (Qmulti_click_delay);
             if (rep_INTP(multi_click_delay))
                 delay = rep_INT(multi_click_delay);
             else
@@ -390,8 +397,7 @@ lookup_binding(u_long code, u_long mods, bool (*callback)(repv key),
 
     if(nkp == rep_NULL || nkp == Qglobal_keymap)
     {
-
-	repv tem = Fsymbol_value (Qoverride_keymap, Qt);
+	repv tem = global_symbol_value (Qoverride_keymap);
 	if (tem != Qnil && !rep_VOIDP(tem))
 	    k = search_keymap (tem, code, mods, callback);
 	else
@@ -410,13 +416,13 @@ lookup_binding(u_long code, u_long mods, bool (*callback)(repv key),
 
 	    if (!k && current_window)
 	    {
-		/* 2. search focused/pointer window keymap property */
+		/* 3. search focused/pointer window keymap property */
 		tem = Fwindow_get (rep_VAL(current_window), Qkeymap);
 		if (tem && tem != Qnil)
 		    k = search_keymap(tem, code, mods, callback);
 	    }
 	    if(!k)
-		/* 3. search global-keymap */
+		/* 4. search global-keymap */
 		k = search_keymap(Qglobal_keymap, code, mods, callback);
 	}
     }
@@ -448,20 +454,9 @@ eval_input_callback(repv key)
 	    {
 		/* An autoload, try to load it. */
 		rep_GC_root gc_key;
-#if rep_INTERFACE >= 9
 		rep_PUSHGC (gc_key, key);
 		cmd = rep_call_with_closure (cmd, rep_load_autoload, cmd);
 		rep_POPGC;
-#else
-		struct rep_Call lc;
-		lc.fun = lc.args = lc.args_evalled_p = Qnil;
-		rep_PUSH_CALL(lc);
-		rep_USE_FUNARG(cmd);
-		rep_PUSHGC(gc_key, key);
-		cmd = rep_load_autoload(cmd);
-		rep_POPGC;
-		rep_POP_CALL(lc);
-#endif
 		if(cmd == rep_NULL)
 		    return FALSE;
 	    }
@@ -485,7 +480,7 @@ static repv
 lookup_event_binding (u_long code, u_long mods, repv context_map)
 {
     Lisp_Window *w = 0;
-    if (current_x_event && mods & EV_TYPE_MOUSE)
+    if (current_x_event && (mods & EV_TYPE_MOUSE))
     {
 	/* If a mouse event, look for bindings in the window that
 	   the button was pressed in, not where the input focus is. */
@@ -523,7 +518,7 @@ eval_input_event(repv context_map)
     if (next_keymap_path == rep_NULL)
     {
 	if (print_prefix)
-	    rep_call_lisp1 (Fsymbol_value (Qdisplay_message, Qt), Qnil);
+	    rep_call_lisp1 (global_symbol_value (Qdisplay_message), Qnil);
 	print_prefix = FALSE;
 	event_index = 0;
     }
@@ -531,7 +526,7 @@ eval_input_event(repv context_map)
     if(cmd != rep_NULL)
     {
 	/* Found a binding for this event; evaluate it. */
-	result = Fcall_command(cmd, Qnil);
+	result = rep_call_lisp1 (Fsymbol_value (Qcall_command, Qt), cmd);
     }
     else if(next_keymap_path != rep_NULL)
     {
@@ -554,7 +549,7 @@ eval_input_event(repv context_map)
     {
 	/* An unbound key with no prefix keys. */
 
-	repv hook = Fsymbol_value (Qunbound_key_hook, Qt);
+	repv hook = global_symbol_value (Qunbound_key_hook);
 	if (!rep_VOIDP (hook) && hook != Qnil)
 	    result = Fcall_hook(Qunbound_key_hook, Qnil, Qor);
 	else if (rep_recurse_depth == 0
@@ -565,8 +560,6 @@ eval_input_event(repv context_map)
 	       we've got stuck with an unbreakable grab somehow.. */
 	    Fungrab_keyboard ();
 	    Fungrab_pointer ();
-	    fprintf (stderr, "yow! may have just avoided a lockup!\n");
-	    Fbeep ();
 	    result = Fthrow (Qtop_level, Qnil);
 	}
     }
@@ -577,8 +570,11 @@ eval_input_event(repv context_map)
 
     last_event[0] = current_event[0];
     last_event[1] = current_event[1];
-    current_event[0] = current_event[1] = 0;
- 
+
+    /* Only lose the current event when in the topmost event loop. */
+    if (rep_recurse_depth == 0)
+	current_event[0] = current_event[1] = 0;
+
     if (print_prefix)
 	print_event_prefix ();
 
@@ -595,28 +591,28 @@ struct key_def {
 
 static struct key_def default_mods[] = {
     { "S",	  ShiftMask },
-    { "shift",	  ShiftMask },
+    { "Shift",	  ShiftMask },
     { "C",        ControlMask },
-    { "control",  ControlMask },
+    { "Control",  ControlMask },
     { "M",        EV_MOD_META },
-    { "meta",     EV_MOD_META },
+    { "Meta",     EV_MOD_META },
     { "A",        EV_MOD_ALT },
-    { "alt",      EV_MOD_ALT },
+    { "Alt",      EV_MOD_ALT },
     { "H",        EV_MOD_HYPER },
-    { "hyper",    EV_MOD_HYPER },
-    { "super",    EV_MOD_SUPER },
-    { "mod1",     Mod1Mask },
-    { "mod2",     Mod2Mask },
-    { "mod3",     Mod3Mask },
-    { "mod4",     Mod4Mask },
-    { "mod5",     Mod5Mask },
-    { "button1",  Button1Mask },
-    { "button2",  Button2Mask },
-    { "button3",  Button3Mask },
-    { "button4",  Button4Mask },
-    { "button5",  Button5Mask },
-    { "any",      EV_MOD_ANY },
-    { "release",  EV_MOD_RELEASE },
+    { "Hyper",    EV_MOD_HYPER },
+    { "Super",    EV_MOD_SUPER },
+    { "Mod1",     Mod1Mask },
+    { "Mod2",     Mod2Mask },
+    { "Mod3",     Mod3Mask },
+    { "Mod4",     Mod4Mask },
+    { "Mod5",     Mod5Mask },
+    { "Button1",  Button1Mask },
+    { "Button2",  Button2Mask },
+    { "Button3",  Button3Mask },
+    { "Button4",  Button4Mask },
+    { "Button5",  Button5Mask },
+    { "Any",      EV_MOD_ANY },
+    { "Release",  EV_MOD_RELEASE },
     { 0, 0 }
 };
 
@@ -723,6 +719,17 @@ lookup_event(u_long *code, u_long *mods, u_char *desc)
 	ks = XStringToKeysym(desc);
 	if(ks != NoSymbol)
 	{
+	    if (*mods & ShiftMask)
+	    {
+		KeySym lower, upper;
+		XConvertCase (ks, &lower, &upper);
+		if (ks == lower && upper != lower)
+		{
+		    /* canonify `S-x' to `X' */
+		    *mods &= ~ShiftMask;
+		    ks = upper;
+		}
+	    }
 	    *mods |= EV_TYPE_KEY;
 	    *code = ks;
 	    return TRUE;
@@ -820,7 +827,7 @@ print_event_prefix(void)
 	*bufp++ = '.';
     }
 
-    rep_call_lisp1 (Fsymbol_value (Qdisplay_message, Qt),
+    rep_call_lisp1 (global_symbol_value (Qdisplay_message),
 		    rep_string_dupn (buf, bufp - buf));
     printed_this_prefix = TRUE;
 
@@ -831,7 +838,7 @@ print_event_prefix(void)
 /* Lisp functions */
 
 DEFUN("make-keymap", Fmake_keymap, Smake_keymap, (void), rep_Subr0) /*
-::doc:make-keymap::
+::doc:sawfish.wm.events#make-keymap::
 make-keymap
 
 Return a new keymap suitable for storing bindings in. This is a cons cell
@@ -843,7 +850,7 @@ nil.
 }
 
 DEFUN("bind-keys", Fbind_keys, Sbind_keys, (repv args), rep_SubrN) /*
-::doc:bind-keys::
+::doc:sawfish.wm.events#bind-keys::
 bind-keys KEYMAP { EVENT-DESCRIPTION COMMAND }...
 
 Adds key bindings to KEYMAP. Each EVENT-DESCRIPTION is a string naming an
@@ -892,7 +899,7 @@ Returns KEYMAP when successful.
 }
 
 DEFUN("unbind-keys", Funbind_keys, Sunbind_keys, (repv args), rep_SubrN) /*
-::doc:unbind-keys::
+::doc:sawfish.wm.events#unbind-keys::
 unbind-keys KEY-MAP EVENT-DESCRIPTION...
 ::end:: */
 {
@@ -947,7 +954,7 @@ unbind-keys KEY-MAP EVENT-DESCRIPTION...
 }
 
 DEFUN("grab-keymap", Fgrab_keymap, Sgrab_keymap, (repv map), rep_Subr1) /*
-::doc:grab-keymap::
+::doc:sawfish.wm.events#grab-keymap::
 grab-keymap KEYMAP
 
 Grab any events in KEYMAP that need to be grabbed so that bindings in
@@ -960,7 +967,7 @@ KEYMAP may be serviced.
 }
 
 DEFUN("ungrab-keymap", Fungrab_keymap, Sungrab_keymap, (repv map), rep_Subr1)/*
-::doc:ungrab-keymap::
+::doc:sawfish.wm.events#ungrab-keymap::
 ungrab-keymap KEYMAP
 
 Ungrab any events in KEYMAP that would have been grabbed so that bindings in
@@ -974,7 +981,7 @@ KEYMAP may be serviced.
 
 DEFSTRING(not_in_handler, "Not in event handler");
 DEFUN("current-event-string", Fcurrent_event_string, Scurrent_event_string, (void), rep_Subr0) /*
-::doc:current-event-string::
+::doc:sawfish.wm.events#current-event-string::
 current-event-string
 
 Returns the string which would have been inserted by the current event if
@@ -997,7 +1004,7 @@ a Lisp function hadn't been called instead.
 }
 
 DEFUN("current-event", Fcurrent_event, Scurrent_event, (void), rep_Subr0) /*
-::doc:current-event::
+::doc:sawfish.wm.events#current-event::
 current-event
 
 Return the event which caused the current command to be invoked.
@@ -1012,7 +1019,7 @@ Return the event which caused the current command to be invoked.
 
 DEFUN("proxy-current-event", Fproxy_current_event, Sproxy_current_event,
       (repv win, repv mask, repv prop), rep_Subr3) /*
-::doc:proxy-current-event::
+::doc:sawfish.wm.events#proxy-current-event::
 proxy-current-event WINDOW [MASK] [PROPAGATE]
 
 Send the current X event to WINDOW, either a window object, a numeric
@@ -1039,7 +1046,7 @@ grab will be released first.
 }
 
 DEFUN("allow-events", Fallow_events, Sallow_events, (repv mode), rep_Subr1) /*
-::doc:allow-events::
+::doc:sawfish.wm.events#allow-events::
 allow-events MODE
 ::end:: */
 {
@@ -1068,7 +1075,7 @@ allow-events MODE
 }
 
 DEFUN("last-event", Flast_event, Slast_event, (void), rep_Subr0) /*
-::doc:last-event::
+::doc:sawfish.wm.events#last-event::
 last-event
 
 Return the previous event which occurred.
@@ -1082,7 +1089,7 @@ Return the previous event which occurred.
 }
 
 DEFUN("event-name", Fevent_name, Sevent_name, (repv ev), rep_Subr1) /*
-::doc:event-name::
+::doc:sawfish.wm.events#event-name::
 event-name EVENT
 
 Returns a string naming the event EVENT.
@@ -1102,7 +1109,7 @@ Returns a string naming the event EVENT.
 }
 
 DEFUN("lookup-event", Flookup_event, Slookup_event, (repv name), rep_Subr1) /*
-::doc:lookup-event::
+::doc:sawfish.wm.events#lookup-event::
 lookup-event EVENT-NAME
 
 Return the event whose name is EVENT-NAME.
@@ -1118,7 +1125,7 @@ Return the event whose name is EVENT-NAME.
 }
 
 DEFUN("lookup-event-binding", Flookup_event_binding, Slookup_event_binding, (repv ev), rep_Subr1) /*
-::doc:lookup-event-binding::
+::doc:sawfish.wm.events#lookup-event-binding::
 lookup-event-binding EVENT
 
 Return the command currently associated with the event EVENT.
@@ -1151,7 +1158,7 @@ currently focused window for keyboard events.
 
 DEFUN("search-keymap", Fsearch_keymap, Ssearch_keymap,
       (repv ev, repv km), rep_Subr2) /*
-::doc:search-keymap::
+::doc:sawfish.wm.events#search-keymap::
 search-keymap EVENT KEYMAP
 
 Return the (COMMAND . EVENT) binding of EVENT in KEYMAP, or nil.
@@ -1166,7 +1173,7 @@ Return the (COMMAND . EVENT) binding of EVENT in KEYMAP, or nil.
 
 DEFUN("x-lookup-keysym", Fx_lookup_keysym,
       Sx_lookup_keysym, (repv name), rep_Subr1) /*
-::doc:x-lookup-keysym::
+::doc:sawfish.wm.events#x-lookup-keysym::
 x-lookup-keysym NAME
 
 Return the X11 keysym (an integer), named by the Lisp symbol NAME.
@@ -1179,7 +1186,7 @@ Return the X11 keysym (an integer), named by the Lisp symbol NAME.
 }
 
 DEFUN("x-keysym-name", Fx_keysym_name, Sx_keysym_name, (repv ks), rep_Subr1) /*
-::doc:x-keysym-name::
+::doc:sawfish.wm.events#x-keysym-name::
 x-keysym-name KEYSYM
 
 Return the Lisp symbol naming the X11 keysym represented by the integer
@@ -1193,7 +1200,7 @@ KEYSYM.
 }
 
 DEFUN("keymapp", Fkeymapp, Skeymapp, (repv arg), rep_Subr1) /*
-::doc:keymapp::
+::doc:sawfish.wm.events#keymapp::
 keymapp ARG
 
 Returns t if ARG can be used as a keymap.
@@ -1203,7 +1210,7 @@ Returns t if ARG can be used as a keymap.
 }
 
 DEFUN("eventp", Feventp, Seventp, (repv arg), rep_Subr1) /*
-::doc:eventp::
+::doc:sawfish.wm.events#eventp::
 eventp ARG
 
 Returns t if the ARG is an input event.
@@ -1219,15 +1226,55 @@ DEFUN("forget-button-press", Fforget_button_press,
     return Qt;
 }
 
+/* Find the bottom-most window below TOP containing (X,Y) that selects
+   for button events */
+static Window
+window_getting_button_event(Window top, int x, int y)
+{
+    Window w = top, w2 = w, child;
+    XWindowAttributes wa;
+
+    /* Find the bottom-most child of W containing (X,Y) */
+    do {
+	XTranslateCoordinates (dpy, w, w2, x, y, &x, &y, &child);
+	w = w2;
+	if (child)
+	    w2 = child;
+    } while (child);
+
+    /* Then work up until TOP is reached, or a window selecting
+       button events is found */
+again:
+    XGetWindowAttributes (dpy, w, &wa);
+    if (w != top
+	&& !(wa.all_event_masks & (ButtonPressMask | ButtonReleaseMask)))
+    {
+	Window d1, *d2, parent;
+	u_int d3;
+	    
+	XQueryTree (dpy, w, &d1, &parent, &d2, &d3);
+	if (d2 != 0)
+	    XFree (d2);
+	if (parent != 0)
+	{
+	    w = parent;
+	    goto again;
+	}
+    }
+
+    return w;
+}
+
 DEFUN("synthesize-event", Fsynthesize_event, Ssynthesize_event,
       (repv event, repv win, repv propagate), rep_Subr3) /*
-::doc:synthesize-event::
+::doc:sawfish.wm.events#synthesize-event::
 synthesize-event EVENT WINDOW [PROPAGATE]
 ::end:: */
 {
     XEvent ev;
     repv ptr = Fquery_pointer (Qnil);
     Window w = x_win_from_arg (win);
+    Window child, dummy;
     int x_offset, y_offset;
 
     if (w == 0)
@@ -1286,16 +1333,24 @@ synthesize-event EVENT WINDOW [PROPAGATE]
 	ev.xbutton.root = root_window;
 	ev.xbutton.subwindow = 0;
 	ev.xbutton.time = last_event_time;
-	ev.xbutton.x_root = rep_INT (rep_CAR (ptr));
-	ev.xbutton.y_root = rep_INT (rep_CDR (ptr));
-	ev.xbutton.x = (ev.xbutton.x_root + x_offset);
-	ev.xbutton.y = (ev.xbutton.y_root + y_offset);
+	ev.xbutton.x = rep_INT (rep_CAR (ptr)) + x_offset;
+	ev.xbutton.y = rep_INT (rep_CDR (ptr)) + y_offset;
+	XTranslateCoordinates (dpy, w, root_window,
+			       ev.xbutton.x, ev.xbutton.y,
+			       &ev.xbutton.x_root, &ev.xbutton.y_root,
+			       &dummy);
+	child = window_getting_button_event (w, ev.xbutton.x, ev.xbutton.y);
+	XTranslateCoordinates (dpy, w, child,
+			       ev.xbutton.x, ev.xbutton.y,
+			       &ev.xbutton.x, &ev.xbutton.y,
+			       &dummy);
 	ev.xbutton.same_screen = True;
+	ev.xbutton.window = child;
 	ev.xany.type = ButtonPress;
-	XSendEvent (dpy, w, propagate != Qnil, ButtonPressMask, &ev);
+	XSendEvent (dpy, child, propagate != Qnil, ButtonPressMask, &ev);
 	ev.xany.type = ButtonRelease;
 	ev.xbutton.state |= mask;
-	XSendEvent (dpy, w, propagate != Qnil, ButtonReleaseMask, &ev);
+	XSendEvent (dpy, child, propagate != Qnil, ButtonReleaseMask, &ev);
 	break;
 
     default:
@@ -1653,6 +1708,8 @@ grab_window_events (Lisp_Window *w, bool grab)
 void
 keys_init(void)
 {
+    repv tem;
+
     rep_INTERN_SPECIAL(global_keymap);
     rep_INTERN_SPECIAL(root_window_keymap);
     rep_INTERN_SPECIAL(override_keymap);
@@ -1663,6 +1720,7 @@ keys_init(void)
     Fset (Qeval_modifier_events, Qnil);
     rep_INTERN(keymap);
 
+    tem = rep_push_structure ("sawfish.wm.events");
     rep_ADD_SUBR(Smake_keymap);
     rep_ADD_SUBR(Sbind_keys);
     rep_ADD_SUBR(Sunbind_keys);
@@ -1683,6 +1741,7 @@ keys_init(void)
     rep_ADD_SUBR(Sx_lookup_keysym);
     rep_ADD_SUBR(Sx_keysym_name);
     rep_ADD_SUBR(Ssynthesize_event);
+    rep_pop_structure (tem);
 
     rep_INTERN(async_pointer);
     rep_INTERN(async_keyboard);
@@ -1704,7 +1763,17 @@ keys_init(void)
     rep_INTERN_SPECIAL(multi_click_delay);
     Fset (Qmulti_click_delay, rep_MAKE_INT(DEFAULT_DOUBLE_CLICK_TIME));
 
+    rep_INTERN_SPECIAL(this_command);
+    rep_INTERN_SPECIAL(last_command);
+    rep_INTERN_SPECIAL(prefix_arg);
+    rep_INTERN_SPECIAL(current_prefix_arg);
+    Fset (Qthis_command, Qnil);
+    Fset (Qlast_command, Qnil);
+    Fset (Qprefix_arg, Qnil);
+    Fset (Qcurrent_prefix_arg, Qnil);
+
     rep_INTERN(display_message);
+    rep_INTERN(call_command);
 
     rep_mark_static(&next_keymap_path);
 
