@@ -1,5 +1,5 @@
 /* sawmill_subrs.h -- prototypes
-   $Id: sawmill_subrs.h,v 1.88 2000/06/07 09:38:01 john Exp $
+   $Id: sawmill_subrs.h,v 1.91 2000/09/17 12:30:08 john Exp $
 
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
@@ -53,9 +53,8 @@ extern void commands_init (void);
 /* from display.c */
 extern char *prog_name;
 extern Display *dpy;
-extern int screen_num, screen_width, screen_height, screen_depth;
-extern Visual *screen_visual;
-extern Colormap screen_cmap;
+extern int screen_num, screen_width, screen_height, preferred_depth;
+extern Visual *preferred_visual;
 extern Window root_window, no_focus_window;
 extern Atom xa_wm_state, xa_wm_change_state, xa_wm_protocols,
     xa_wm_delete_window, xa_wm_colormap_windows, xa_wm_take_focus,
@@ -131,6 +130,8 @@ extern repv Qcursor, Qfocused, Qhighlighted, Qclicked;
 extern bool frame_state_mutex;
 extern int current_state (struct frame_part *fp);
 extern void set_frame_shapes (Lisp_Window *w, bool atomic);
+extern void queue_reshape_frame (Lisp_Window *w);
+extern void commit_queued_reshapes (void);
 extern void refresh_frame_part (struct frame_part *fp);
 extern void refresh_frame_parts (Lisp_Window *w);
 extern void create_window_frame (Lisp_Window *w);
@@ -258,6 +259,7 @@ extern int main (int argc, char **argv);
 extern repv Fquit (void);
 extern repv Frestart (void);
 extern void add_hook (repv sym, repv fun);
+extern repv module_symbol_value (repv mod, repv sym);
 extern repv global_symbol_value (repv sym);
 
 /* from pixmap-cache.c */
