@@ -1,5 +1,5 @@
 /* windows.c -- window manipulation
-   $Id: windows.c,v 1.89 2000/06/07 18:11:34 john Exp $
+   $Id: windows.c,v 1.104 2000/11/13 19:36:55 jsh Exp $
 
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
@@ -38,6 +38,7 @@ DEFSYM(place_window_hook, "place-window-hook");
 DEFSYM(placed, "placed");
 DEFSYM(after_framing_hook, "after-framing-hook");
 DEFSYM(after_initialization_hook, "after-initialization-hook");
+DEFSYM(remove_window_hook, "remove-window-hook");
 
 /* for visibility-notify-hook */
 DEFSYM(fully_obscured, "fully-obscured");
@@ -89,6 +90,16 @@ mapped_not_override_p (Window id)
     return ((wa.map_state != IsUnmapped) && (wa.override_redirect != True));
 }
 
+/* Returns true if the window's Input hint is set (or defaults to being set) */
+static bool
+window_input_hint_p (Lisp_Window *w)
+{
+    if (w->wmhints != 0 && (w->wmhints->flags & InputHint))
+	return w->wmhints->input;
+    else
+	return TRUE;
+}
+
 /* Give the input focus to window W, or to no window if W is null */
 void
 focus_on_window (Lisp_Window *w)
@@ -98,18 +109,28 @@ focus_on_window (Lisp_Window *w)
 	Window focus;
 	DB(("focus_on_window (%s)\n", rep_STR(w->name)));
 	if (!w->client_unmapped)
-	{
+	{                              
 	    if (w->does_wm_take_focus)
 	    {
 		DB(("  sending WM_TAKE_FOCUS message\n"));
 		send_client_message (w->id, xa_wm_take_focus, last_event_time);
+
+		/* Only focus on the window if accepts-input is true */
+		if (window_input_hint_p (w))
+		    focus = w->id;
+		else
+		    focus = 0;
 	    }
-	    focus = w->id;
+	    else
+		focus = w->id;
 	}
 	else
 	    focus = w->frame;
-	XSetInputFocus (dpy, focus, RevertToParent, last_event_time);
-	focus_window = w;
+	if (focus != 0)
+	{
+	    XSetInputFocus (dpy, focus, RevertToParent, last_event_time);
+	    focus_window = w;
+	}
     }
     else
     {
@@ -278,6 +299,7 @@ add_window (Window id)
 	w->saved_id = id;
 	w->plist = Qnil;
 	w->frame_style = Qnil;;
+	w->icon_image = rep_NULL;
 
 	/* ..now do the X11 stuff */
 
@@ -352,6 +374,9 @@ add_window (Window id)
         w->visible = TRUE;
 	w->mapped = TRUE;		/* only called from map request */
 
+	if (initialising)
+	    Fwindow_put (rep_VAL (w), Qplaced, Qt);
+
 	/* ..then call the add-window-hook's.. */
 	rep_PUSHGC(gc_win, win);
 	Fcall_window_hook (Qbefore_add_window_hook, rep_VAL(w), Qnil, Qnil);
@@ -367,18 +392,18 @@ add_window (Window id)
 	    create_window_frame (w);
 	    install_window_frame (w);
 
-           /* this grabs bound events in the subwindow */
-           grab_window_events (w, TRUE);
+	    /* this grabs bound events in the subwindow */
+	    grab_window_events (w, TRUE);
 
 	    Fungrab_server ();
 	}
 	else
 	    emit_pending_destroys ();
 
-	if (w->id != 0 && !initialising)
+	if (w->id != 0)
 	{
 	    repv tem = Fwindow_get (rep_VAL(w), Qplaced);
-	    if (tem && tem == Qnil)
+	    if (initialising || (tem && tem == Qnil))
 	    {
 		/* ..then the place-window-hook.. */
 		rep_PUSHGC(gc_win, win);
@@ -417,6 +442,9 @@ remove_window (Lisp_Window *w, repv destroyed, repv from_error)
 	{
 	    grab_window_events (w, FALSE);
 	    remove_window_frame (w);
+
+	    /* Restore original border width of the client */
+	    XSetWindowBorderWidth (dpy, w->id, w->attr.border_width);
 	}
 
 	if (from_error == Qnil)
@@ -470,7 +498,6 @@ emit_pending_destroys (void)
 	{
 	    if (w->id == 0 && !w->destroyed)
 	    {
-		pending_destroys--;
 		w->destroyed = 1;
 		Fcall_window_hook (Qdestroy_notify_hook,
 				   rep_VAL(w), Qnil, Qnil);
@@ -484,6 +511,7 @@ emit_pending_destroys (void)
 	    }
 	}
     }
+    pending_destroys = 0;
 }
 
 
@@ -491,7 +519,7 @@ emit_pending_destroys (void)
 
 DEFUN("window-get", Fwindow_get, Swindow_get,
       (repv win, repv prop), rep_Subr2) /*
-::doc::window-get::
+::doc:sawfish.wm.windows.subrs#window-get::
 window-get WINDOW PROPERTY
 
 Return the value of the property named PROPERTY (a symbol) of WINDOW.
@@ -517,7 +545,7 @@ Note that these are Lisp properties not X properties.
 
 DEFUN("map-window-properties", Fmap_window_properties,
       Smap_window_properties, (repv fun, repv win), rep_Subr2) /*
-::doc:map-window-properties::
+::doc:sawfish.wm.windows.subrs#map-window-properties::
 map-window-properties FUNCTION WINDOW
 
 Call (FUNCTION PROPERTY VALUE) for all Lisp properties set on window
@@ -543,7 +571,7 @@ object WINDOW.
 
 DEFUN("window-put", Fwindow_put, Swindow_put,
       (repv win, repv prop, repv val), rep_Subr3) /*
-::doc:window-put::
+::doc:sawfish.wm.windows.subrs#window-put::
 window-put WINDOW PROPERTY VALUE
 
 Set the value of the property named PROPERTY (a symbol) of WINDOW to VALUE.
@@ -566,6 +594,7 @@ Note that these are Lisp properties not X properties.
 		grab_keymap_events (VWIN(win)->id,
 				    rep_CAR(rep_CDR(plist)), FALSE);
 		grab_keymap_events (VWIN(win)->id, val, TRUE);
+
 	    }
 	    rep_CAR(rep_CDR(plist)) = val;
 	    return val;
@@ -579,7 +608,7 @@ Note that these are Lisp properties not X properties.
 }
 
 DEFUN("window-name", Fwindow_name, Swindow_name, (repv win), rep_Subr1) /*
-::doc:window-name::
+::doc:sawfish.wm.windows.subrs#window-name::
 window-name WINDOW
 
 Return the name of window object WINDOW.
@@ -591,7 +620,7 @@ Return the name of window object WINDOW.
 
 DEFUN("window-full-name", Fwindow_full_name, Swindow_full_name,
       (repv win), rep_Subr1) /*
-::doc:window-full-name::
+::doc:sawfish.wm.windows.subrs#window-full-name::
 window-full-name WINDOW
 
 Return the full name of window object WINDOW.
@@ -603,7 +632,7 @@ Return the full name of window object WINDOW.
 
 DEFUN("window-icon-name", Fwindow_icon_name, Swindow_icon_name,
       (repv win), rep_Subr1) /*
-::doc:window-icon-name::
+::doc:sawfish.wm.windows.subrs#window-icon-name::
 window-icon-name WINDOW
 
 Return the name of window object WINDOW's icon.
@@ -615,7 +644,7 @@ Return the name of window object WINDOW's icon.
 
 DEFUN("window-mapped-p", Fwindow_mapped_p, Swindow_mapped_p,
       (repv win), rep_Subr1) /*
-::doc:window-mapped-p::
+::doc:sawfish.wm.windows.subrs#window-mapped-p::
 window-mapped-p WINDOW
 
 Return t if the client window associated with object WINDOW is mapped.
@@ -627,7 +656,7 @@ Return t if the client window associated with object WINDOW is mapped.
 }
 
 DEFUN("window-frame", Fwindow_frame, Swindow_frame, (repv win), rep_Subr1) /*
-::doc:window-frame::
+::doc:sawfish.wm.windows.subrs#window-frame::
 window-frame WINDOW
 
 Return the frame object associated with WINDOW.
@@ -639,7 +668,7 @@ Return the frame object associated with WINDOW.
 
 DEFUN("set-window-frame", Fset_window_frame, Sset_window_frame,
       (repv win, repv frame), rep_Subr2) /*
-::doc:set-window-frame::
+::doc:sawfish.wm.windows.subrs#set-window-frame::
 set-window-frame WINDOW FRAME
 
 Set the frame associated with the window object WINDOW to FRAME (a
@@ -665,7 +694,7 @@ new frame constructed as specified by FRAME.
 }
 
 DEFUN("rebuild-frame", Frebuild_frame, Srebuild_frame, (repv win), rep_Subr1) /*
-::doc:rebuild-frame::
+::doc:sawfish.wm.windows.subrs#rebuild-frame::
 rebuild-frame WINDOW
 
 Reinitialises and recalibrates the window frame of WINDOW.
@@ -683,7 +712,7 @@ Reinitialises and recalibrates the window frame of WINDOW.
 
 DEFUN("window-position", Fwindow_position, Swindow_position,
       (repv win), rep_Subr1) /*
-::doc:window-position::
+::doc:sawfish.wm.windows.subrs#window-position::
 window-position WINDOW
 
 Return (X . Y) defining the current position of WINDOW.
@@ -696,7 +725,7 @@ Return (X . Y) defining the current position of WINDOW.
 
 DEFUN("window-dimensions", Fwindow_dimensions, Swindow_dimensions,
       (repv win), rep_Subr1) /*
-::doc:window-dimensions::
+::doc:sawfish.wm.windows.subrs#window-dimensions::
 window-dimensions WINDOW
 
 Return (WIDTH . HEIGHT) defining the current dimensions of the client
@@ -710,7 +739,7 @@ window associated with WINDOW.
 
 DEFUN("window-frame-dimensions", Fwindow_frame_dimensions,
       Swindow_frame_dimensions, (repv win), rep_Subr1) /*
-::doc:window-frame-dimensions::
+::doc:sawfish.wm.windows.subrs#window-frame-dimensions::
 window-frame-dimensions WINDOW
 
 Return (WIDTH . HEIGHT) defining the current dimensions of the frame
@@ -729,7 +758,7 @@ surrounding WINDOW.
 
 DEFUN("window-frame-offset", Fwindow_frame_offset,
       Swindow_frame_offset, (repv win), rep_Subr1) /*
-::doc:window-frame-offset::
+::doc:sawfish.wm.windows.subrs#window-frame-offset::
 window-frame-offset WINDOW
 
 Return (X . Y) defining the offset from the origin of the client window
@@ -742,7 +771,7 @@ associated with WINDOW to its frame window.
 }
 
 DEFUN("windowp", Fwindowp, Swindowp, (repv win), rep_Subr1) /*
-::doc:windowp::
+::doc:sawfish.wm.windows.subrs#windowp::
 windowp ARG
 
 Return t if ARG is a window object.
@@ -753,14 +782,14 @@ Return t if ARG is a window object.
 
 DEFUN("set-input-focus", Fset_input_focus, Sset_input_focus,
       (repv win), rep_Subr1) /*
-::doc:set-input-focus::
+::doc:sawfish.wm.windows.subrs#set-input-focus::
 set-input-focus WINDOW
 
 Set the input focus to WINDOW. If WINDOW is nil, then no window will
 have the focus.
 ::end:: */
 {
-    if (win != Qnil)
+    if (win != Qnil && win != Qroot)
     {
 	rep_DECLARE1(win, WINDOWP);
 	focus_on_window (VWIN(win));
@@ -771,7 +800,7 @@ have the focus.
 }
 
 DEFUN("input-focus", Finput_focus, Sinput_focus, (void), rep_Subr0) /*
-::doc:input-focus::
+::doc:sawfish.wm.windows.subrs#input-focus::
 input-focus
 
 Return the window object that has the input focus, or nil if none does.
@@ -782,7 +811,7 @@ Return the window object that has the input focus, or nil if none does.
 
 DEFUN("window-wants-input-p", Fwindow_wants_input_p, Swindow_wants_input_p,
       (repv win), rep_Subr1) /*
-::doc:window-wants-input-p::
+::doc:sawfish.wm.windows.subrs#window-wants-input-p::
 window-wants-input-p WINDOW
 
 Return t if the client window associated with object WINDOW has hinted
@@ -790,12 +819,15 @@ that it would like to be given the input focus when applicable.
 ::end:: */
 {
     rep_DECLARE1(win, WINDOWP);
-    return (VWIN(win)->wmhints && VWIN(win)->wmhints->input) ? Qt : Qnil;
+    if (VWIN(win)->does_wm_take_focus)
+	return Qt;
+    else
+	return window_input_hint_p (VWIN (win)) ? Qt : Qnil;
 }
 
 DEFUN("managed-windows", Fmanaged_windows, Smanaged_windows,
       (void), rep_Subr0) /*
-::doc:managed-windows::
+::doc:sawfish.wm.windows.subrs#managed-windows::
 managed-windows
 
 Return a list of all known client window objects.
@@ -814,7 +846,7 @@ Return a list of all known client window objects.
 
 DEFUN("get-window-by-id", Fget_window_by_id, Sget_window_by_id,
       (repv id), rep_Subr1) /*
-::doc:get-window-by-id::
+::doc:sawfish.wm.windows.subrs#get-window-by-id::
 get-window-by-id ID
 
 Return the window object associated with xid ID, or nil.
@@ -827,7 +859,7 @@ Return the window object associated with xid ID, or nil.
 }
 
 DEFUN("stacking-order", Fstacking_order, Sstacking_order, (void), rep_Subr0) /*
-::doc:stacking-order::
+::doc:sawfish.wm.windows.subrs#stacking-order::
 stacking-order
 
 Return a list of windows defining the current stacking order of all
@@ -860,7 +892,7 @@ client windows.
 
 DEFUN("window-visibility", Fwindow_visibility, Swindow_visibility,
       (repv win), rep_Subr1) /*
-::doc:window-visibility::
+::doc:sawfish.wm.windows.subrs#window-visibility::
 window-visibility WINDOW
 
 Return a symbol defining the visibility of WINDOW. Possible returned
@@ -887,7 +919,7 @@ symbols are `fully-obscured', `partially-obscured' or `unobscured'.
 
 DEFUN("window-transient-p", Fwindow_transient_p, Swindow_transient_p,
       (repv win), rep_Subr1) /*
-::doc:window-transient-p::
+::doc:sawfish.wm.windows.subrs#window-transient-p::
 window-transient-p WINDOW
 
 Return non-nil if WINDOW is a transient window. The returned value will
@@ -899,9 +931,23 @@ then be the numeric id of its parent window.
 	    ? rep_MAKE_INT(VWIN(win)->transient_for_hint) : Qnil);
 }
 
+DEFUN("window-urgent-p", Fwindow_urgent_p, Swindow_urgent_p,
+      (repv win), rep_Subr1) /*
+::doc:sawfish.wm.windows.subrs#window-urgent-p::
+window-urgent-p WINDOW
+
+Return true if the `Urgency' hint of the window associated with WINDOW
+is set.
+::end:: */
+{
+    rep_DECLARE1 (win, WINDOWP);
+    return ((VWIN (win)->wmhints
+	     && VWIN (win)->wmhints->flags & XUrgencyHint) ? Qt : Qnil);
+}
+
 DEFUN("window-shaped-p", Fwindow_shaped_p, Swindow_shaped_p,
       (repv win), rep_Subr1) /*
-::doc:window-shaped-p::
+::doc:sawfish.wm.windows.subrs#window-shaped-p::
 window-shaped-p WINDOW
 
 Return non-nil if WINDOW is shaped.
@@ -912,7 +958,7 @@ Return non-nil if WINDOW is shaped.
 }
 
 DEFUN("hide-window", Fhide_window, Shide_window, (repv win), rep_Subr1) /*
-::doc:hide-window::
+::doc:sawfish.wm.windows.subrs#hide-window::
 hide-window WINDOW
 
 Prevent WINDOW from being displayed. See `show-window'.
@@ -940,7 +986,7 @@ Prevent WINDOW from being displayed. See `show-window'.
 }
 
 DEFUN("show-window", Fshow_window, Sshow_window, (repv win), rep_Subr1) /*
-::doc:show-window::
+::doc:sawfish.wm.windows.subrs#show-window::
 show-window WINDOW
 
 Ensure that WINDOW (if it has been mapped) is visible. See `hide-window'.
@@ -968,7 +1014,7 @@ Ensure that WINDOW (if it has been mapped) is visible. See `hide-window'.
 
 DEFUN("window-visible-p", Fwindow_visible_p, Swindow_visible_p,
       (repv win), rep_Subr1) /*
-::doc:window-visible-p::
+::doc:sawfish.wm.windows.subrs#window-visible-p::
 window-visible-p WINDOW
 
 Return t if WINDOW is currently visible (i.e. not hidden, see `hide-window').
@@ -980,7 +1026,7 @@ Return t if WINDOW is currently visible (i.e. not hidden, see `hide-window').
 
 DEFUN("window-framed-p", Fwindow_framed_p, Swindow_framed_p,
       (repv win), rep_Subr1) /*
-::doc:window-framed-p::
+::doc:sawfish.wm.windows.subrs#window-framed-p::
 window-framed-p WINDOW
 
 Return t if WINDOW has been reparented to a frame window.
@@ -991,7 +1037,7 @@ Return t if WINDOW has been reparented to a frame window.
 }
 
 DEFUN("window-id", Fwindow_id, Swindow_id, (repv win), rep_Subr1) /*
-::doc:window-id::
+::doc:sawfish.wm.windows.subrs#window-id::
 window-id WINDOW
 
 Return the numeric id of the client window associated with object
@@ -1004,7 +1050,7 @@ WINDOW. Returns nil if the client window has been deleted.
 
 DEFUN("window-group-id", Fwindow_group_id, Swindow_group_id,
       (repv win), rep_Subr1) /*
-::doc:window-group-id::
+::doc:sawfish.wm.windows.subrs#window-group-id::
 window-group-id WINDOW
 
 Return the numeric id defining the leader of the group that WINDOW is a
@@ -1021,7 +1067,7 @@ member of, or nil if it is not a member of a group.
 
 DEFUN("window-border-width", Fwindow_border_width, Swindow_border_width,
       (repv win), rep_Subr1) /*
-::doc:window-border-width::
+::doc:sawfish.wm.windows.subrs#window-border-width::
 window-border-width WINDOW
 ::end:: */
 {
@@ -1031,7 +1077,7 @@ window-border-width WINDOW
 
 DEFUN("window-size-hints", Fwindow_size_hints, Swindow_size_hints,
       (repv win), rep_Subr1) /*
-::doc:window-size-hints::
+::doc:sawfish.wm.windows.subrs#window-size-hints::
 window-size-hints WINDOW
 
 Return an alist defining the size-hints specified by the client window
@@ -1118,7 +1164,7 @@ associated with WINDOW. Possible keys in the alist are `min-height',
 
 DEFUN("call-window-hook", Fcall_window_hook, Scall_window_hook,
       (repv hook, repv win, repv args, repv type), rep_Subr4) /*
-::doc:call-window-hook::
+::doc:sawfish.wm.windows.subrs#call-window-hook::
 call-window-hook HOOK WINDOW &optional ARGS HOOK-TYPE
 
 Call HOOK for WINDOW with extra arguments ARGS. See `call-hook' for a
@@ -1149,6 +1195,113 @@ out:
     return tem;
 }
 
+DEFUN("window-icon-image", Fwindow_icon_image,
+      Swindow_icon_image, (repv win), rep_Subr1) /*
+::doc:sawfish.wm.windows.subrs#window-icon-image::
+window-icon-image WINDOW
+
+Returns an image object representing the icon currently associated with
+WINDOW. Returns the symbol `nil' if no such image.
+::end:: */
+{
+   rep_DECLARE1 (win, WINDOWP);
+
+   if (VWIN (win)->icon_image == rep_NULL)
+   {
+       if (VWIN (win)->wmhints == 0)
+	   VWIN (win)->icon_image = Qnil;
+       else
+       {
+	   repv id, mask_id;
+	   repv ret = Qnil;
+
+	   if (VWIN (win)->wmhints->flags & IconPixmapHint
+	       && VWIN (win)->wmhints->icon_pixmap != 0)
+	   {
+	       id = rep_MAKE_INT (VWIN (win)->wmhints->icon_pixmap);
+	   }
+	   else
+	       id = Qnil;
+
+	   if (VWIN (win)->wmhints->flags & IconMaskHint
+	       && VWIN (win)->wmhints->icon_mask != 0)
+	   {
+	       mask_id = rep_MAKE_INT (VWIN (win)->wmhints->icon_mask);
+	   }
+	   else
+	       mask_id = Qnil;
+
+	   if (id != Qnil)
+	       ret = Fmake_image_from_x_drawable (id, mask_id);
+
+	   VWIN (win)->icon_image = ret;
+       }
+   }
+
+   return VWIN (win)->icon_image;
+}
+
+DEFUN ("map-windows", Fmap_windows, Smap_windows, (repv fun), rep_Subr1) /*
+::doc:sawfish.wm.windows.subrs#map-windows::
+map-windows FUN
+
+Map the single-parameter function FUN over all existing windows.
+::end:: */
+{
+    repv w;
+    rep_GC_root gc_fun, gc_w;
+    repv ret = Qnil;
+
+    rep_PUSHGC (gc_fun, fun);
+    rep_PUSHGC (gc_w, w);
+    for (w = rep_VAL (window_list); w != rep_NULL; w = rep_VAL (VWIN(w)->next))
+    {
+	if (VWIN (w)->id != 0)
+	{
+	    ret = rep_call_lisp1 (fun, w);
+	    if (ret == rep_NULL)
+		break;
+	}
+    }
+    rep_POPGC; rep_POPGC;
+    return ret;
+}
+
+DEFUN ("filter-windows", Ffilter_windows,
+       Sfilter_windows, (repv pred), rep_Subr1) /*
+::doc:sawfish.wm.windows.subrs#filter-windows::
+filter-windows PRED
+
+Return the list of windows that match the predicate function PRED.
+::end:: */
+{
+    repv w, output = Qnil, *ptr = &output;
+    rep_GC_root gc_pred, gc_w, gc_output;
+
+    rep_PUSHGC(gc_pred, pred);
+    rep_PUSHGC(gc_w, w);
+    rep_PUSHGC(gc_output, output);
+    for (w = rep_VAL (window_list); w != rep_NULL; w = rep_VAL (VWIN(w)->next))
+    {
+	if (VWIN (w)->id != 0)
+	{
+	    repv tem = rep_call_lisp1 (pred, w);
+	    if (tem == rep_NULL)
+	    {
+		output = rep_NULL;
+		break;
+	    }
+	    if (tem != Qnil)
+	    {
+		*ptr = Fcons (w, Qnil);
+		ptr = rep_CDRLOC (*ptr);
+	    }
+	}
+    }
+    rep_POPGC; rep_POPGC; rep_POPGC;
+    return output;
+}
+
 
 /* type hooks */
 
@@ -1175,6 +1328,7 @@ window_mark (repv win)
     rep_MARKVAL(VWIN(win)->name);
     rep_MARKVAL(VWIN(win)->full_name);
     rep_MARKVAL(VWIN(win)->icon_name);
+    rep_MARKVAL(VWIN(win)->icon_image);
 }
 
 static void
@@ -1227,6 +1381,23 @@ manage_windows (void)
     Fgrab_server ();
 
     XGetInputFocus (dpy, &focus, &revert_to);
+    if (focus == PointerRoot)
+    {
+    	Window root, child;
+	Bool found;
+	int rx, ry, wx, wy;
+	unsigned mask;
+
+	found = XQueryPointer (dpy, DefaultRootWindow(dpy), &root,
+			       &child, &rx, &ry, &wx, &wy, &mask);
+	if (!found)
+	{
+	    found = XQueryPointer (dpy, root, &root, &child,
+				   &rx, &ry, &wx, &wy, &mask);
+	}
+	focus = child;
+    }
+
     XQueryTree (dpy, root_window, &root, &parent, &children, &nchildren);
     initialising = TRUE;
     for (i = 0; i < nchildren; i++)
@@ -1246,7 +1417,7 @@ manage_windows (void)
 
     /* Try to keep the current focus state. */
     focus_on_window (0);
-    if (focus != None && focus != PointerRoot)
+    if (focus != None)
     {
 	Lisp_Window *w = find_window_by_id (focus);
 	if (w != 0)
@@ -1260,10 +1431,13 @@ manage_windows (void)
 void
 windows_init (void)
 {
+    repv tem;
     window_type = rep_register_new_type ("window", window_cmp, window_prin,
 					 window_prin, window_sweep,
 					 window_mark, window_mark_type,
 					 0, 0, 0, 0, 0, 0);
+
+    tem = rep_push_structure ("sawfish.wm.windows.subrs");
     rep_ADD_SUBR(Swindow_get);
     rep_ADD_SUBR(Smap_window_properties);
     rep_ADD_SUBR(Swindow_put);
@@ -1287,6 +1461,7 @@ windows_init (void)
     rep_ADD_SUBR(Sstacking_order);
     rep_ADD_SUBR(Swindow_visibility);
     rep_ADD_SUBR(Swindow_transient_p);
+    rep_ADD_SUBR(Swindow_urgent_p);
     rep_ADD_SUBR(Swindow_shaped_p);
     rep_ADD_SUBR(Shide_window);
     rep_ADD_SUBR(Sshow_window);
@@ -1297,6 +1472,10 @@ windows_init (void)
     rep_ADD_SUBR(Swindow_size_hints);
     rep_ADD_SUBR(Scall_window_hook);
     rep_ADD_SUBR(Swindow_border_width);
+    rep_ADD_SUBR(Swindow_icon_image);
+    rep_ADD_SUBR(Smap_windows);
+    rep_ADD_SUBR(Sfilter_windows);
+    rep_pop_structure (tem);
 
     rep_INTERN_SPECIAL(before_add_window_hook);
     rep_INTERN_SPECIAL(add_window_hook);
@@ -1305,6 +1484,7 @@ windows_init (void)
     rep_INTERN(placed);
     rep_INTERN_SPECIAL(after_framing_hook);
     rep_INTERN_SPECIAL(after_initialization_hook);
+    rep_INTERN_SPECIAL(remove_window_hook);
 
     rep_INTERN(fully_obscured);
     rep_INTERN(partially_obscured);
@@ -1354,9 +1534,14 @@ void
 windows_kill (void)
 {
     Lisp_Window *w = window_list;
+    repv next;
+    rep_GC_root gc_next;
+    rep_PUSHGC (gc_next, next);
     while (w != 0)
     {
+	next = rep_VAL (w->next);
+	Fcall_window_hook (Qremove_window_hook, rep_VAL (w), Qnil, Qnil);
 	remove_window (w, Qnil, Qnil);
-	w = w->next;
+	w = VWIN (next);
     }
 }
